#!/usr/bin/env python
"""
This workflow demonstrates execution of a sequence template
with task output redirection

"""
from __future__ import print_function

import os
import sys

import tigres


def check_status():
    """
    Are there any failed tasks?
    """
    task_check = tigres.check('task', state=tigres.utils.State.FAIL)

    if not task_check:
        results = tigres.get_results()
        print(results)
    else:
        raise tigres.utils.TigresException("A Task Failure was found")


def add_up():
    """
    Add up the numbers
    """
    total = 0
    count = 0

    # Read from standard input
    import fileinput
    for line in fileinput.input():
        total += float(line)
        count += 1

    # NOTE: returning a value and printing to standard
    # output will both be captured as output.
    # (e.g. sum:3360.0 count:85)
    print("sum:{}".format(total), end=" ")
    return "count:{}".format(count)


def main():
    tigres.start(name="Example Workflow Sequence Pipeline", log_dest='example_workflow_sequence_redirect.log')
    failed = 0
    try:
        ###################
        # Create the Tasks
        ###################

        # All of the below tasks are able to read from
        # standard input

        # Executable Task Implementation
        task_grep = tigres.Task("grep", tigres.EXECUTABLE, "grep")
        task_awk = tigres.Task("awk", tigres.EXECUTABLE, "awk")
        task_sort = tigres.Task("sort", tigres.EXECUTABLE, "sort")
        task_unique = tigres.Task("uniq", tigres.EXECUTABLE, "uniq")

        # Function task implementation
        task_add_up = tigres.Task("add_up", tigres.FUNCTION, add_up)

        #######################################
        # Sequence with only executable tasks
        #######################################

        # grep write data.csv | awk -F, '{{print $2}}' | sort | uniq

        print("\nSequence Template Redirection\n(execution task implementation)")
        inputs_grep = tigres.InputValues("inputs_grep", ["write", "{}/data.csv".format(os.getcwd())])

        inputs_awk = tigres.InputValues("inputs_awk",
                                        ["-F,", "'{{print $2}}'".format(os.getcwd())])

        inputs_sort = tigres.InputValues("inputs_sort",
                                         [])

        inputs_uniq = tigres.InputValues("inputs_uniq",
                                         [])
        output = tigres.sequence("Pipeline Sequence (executable only)", tigres.TaskArray(name="Pipeline Executables",
                                                                                         tasks=[task_grep, task_awk,
                                                                                                task_sort,
                                                                                                task_unique]),
                                 tigres.InputArray(values=[inputs_grep, inputs_awk, inputs_sort, inputs_uniq]),
                                 redirect=True)

        check_status()
        assert (output == "server-6204\nserver-6401")

        ##############################################
        # Sequence with executable and function tasks
        ##############################################

        # grep write data.csv | awk -F, '{{print $5}}' | add_up()

        # Use all the same inputs except change the column of data
        inputs_awk = tigres.InputValues("inputs_awk",
                                        ["-F,", "'{{print $5}}'".format(os.getcwd())])
        # add_up() takes not inputs
        inputs_add_up = tigres.InputValues("inputs_add_up",
                                           [])

        print("\nPipeline Template Redirection\n(execution and function task implementation)")
        output = tigres.sequence("Pipeline Sequence Mixed", tigres.TaskArray(name="Pipeline Executables",
                                                                             tasks=[task_grep, task_awk, task_add_up]),
                                 tigres.InputArray(values=[inputs_grep, inputs_awk, inputs_add_up]),
                                 redirect=True)
        check_status()
        assert (output == "sum:3360.0 count:85")

    except tigres.utils.TigresException as e:
        failed = 1
        print(e)

    #####################
    # Check for Failures
    #####################
    task_check = tigres.check('task', state=tigres.utils.State.FAIL)
    if task_check:
        # clean up files
        failed = 1
        # There was a failure. Print the tasks that failed
        print("Failed!!", file=sys.stderr)
        for task_record in task_check:
            print(".. Task Failed: {}, {}".format(task_record.name, task_record.errmsg), file=sys.stderr)
    elif not failed:
        print("Success!!")

    # Print the Graph
    # tigres.dot_execution()

    # ################################
    # End the Workflow
    # ################################
    tigres.end()

    if failed:
        raise
    return failed


if __name__ == "__main__":

    main()
