#!/usr/bin/env python
"""
Tigres Split Template example.

This example takes a data files, extracts the data
and performs basic statistics on some of the data (average, median, total) using
a Tigres split template.
"""

import os
import sys
import csv
import collections
import math

import tigres
import statistics



def extract_data(filename, has_header=True):
    """
    Get the data from the csv file
    :param filename:
    :return: list of lists
    :rtype: list
    """
    column_headers = None
    data = []
    count = 0
    with open(filename, 'rU') as csvfile:
        reader = csv.reader(csvfile, delimiter=',', quotechar='"')
        for values in reader:
            if not column_headers and has_header:
                column_headers = values
            else:
                data.append(values)
                count += 1

    return data, column_headers, count


def get_column_data(data, column_num, column_names):
    values = []
    for value in file_col_generator(data, column_num):
        values.append(value)

    return values, column_names[column_num]


def file_col_generator(data, col):
    """
    Generator for the values in the data object for the given column
    :param data:
    :type data: list
    :param col:
    :type col: int
    """
    for values in data:
        value = values[col]
        yield value


def count_categories(values):
    tigres.log_debug("count_categories", message="{}".format(values[1]))

    categories = {}
    for v in values[0]:
        categories.setdefault(v, 0)
        categories[v] += 1

    return categories, values[1]


def total_by_column(data_and_column):
    """
    Calculate the total of the given numeric values in the specified column

    :param data:
    :type data: list
    :param column_num:
    :type column_num: int
    :param column_names: the list of column names
    :type column_names: list
    :return: (operation, column_name, total)
    :rtype: tuple
    """
    return total_by_column.__name__, data_and_column[1], sum(data_and_column[0].values())


def average_by_column(data_and_column):
    """
    Calculate the average of the given numeric values in the specified column

    :param data:
    :type data: list
    :param column_num:
    :type column_num: int
    :param column_names: the list of column names
    :type column_names: list
    :return: (operation, column_name, average)
    :rtype: tuple
    """

    total = sum(data_and_column[0].values())
    count = len(data_and_column[0].values())
    average = total / count
    return average_by_column.__name__, data_and_column[1], average


def median_by_column(data_and_column):
    """
    Calculate the median of the given numeric values in the specified column

    :param data:
    :type data: list
    :param column_num:
    :type column_num: int
    :param column_names: the list of column names
    :type column_names: list
    :return: (operation, column_name, median)
    :rtype: tuple
    """

    return median_by_column.__name__, data_and_column[1], statistics.median(data_and_column[0].values())


def stdev_by_column(data_and_column):

    return stdev_by_column.__name__, data_and_column[1], math.sqrt(statistics.variance(data_and_column[0].values()))


def variance_by_column(data_and_column):

    return variance_by_column.__name__, data_and_column[1], statistics.variance(data_and_column[0].values())


def main(execution, data_file, columns):
    # Initialize the Tigres program
    tigres.start(name="Basic Counts By Column and Category", log_dest=os.path.splitext(__file__)[0] + '.log',
                 execution=execution)

    tigres.set_log_level(tigres.Level.DEBUG)

    # Initialize the field values that we are interested in
    input_cols = [int(c) for c in columns.split(',')]

    # Initialize the operations we would like to perform on the data columns
    impls = [average_by_column, total_by_column, median_by_column, stdev_by_column, variance_by_column]

    # Data Extraction Task
    task_data_extraction = tigres.Task("Data Extraction", tigres.FUNCTION, impl_name=extract_data, input_types=[str])
    inputs_data_extraction = tigres.InputValues("Data File", [data_file])

    input_data = []
    tasks_col_data = []
    tasks_col_categories = []
    input_col_data = []
    inputs_statistics = []
    tasks_statistics = []
    i = 0
    for col in input_cols:
        input_data.append(tigres.InputValues('Column', [tigres.PREVIOUS.task_data_extraction.i[0],
                                                        col, tigres.PREVIOUS.task_data_extraction.i[1]]))
        task_name = "Task.get_column_data.%s" % (col)
        tasks_col_data.append(
            tigres.Task(task_name, task_type=tigres.FUNCTION, impl_name=get_column_data))

        task_catgetory_name = "Task.count_categories.%s" % col
        input_col_data.append(tigres.InputValues('Col Data', [tigres.PREVIOUS.i]))
        tasks_col_categories.append(tigres.Task(task_catgetory_name, tigres.FUNCTION, count_categories))

        for impl in impls:
            # Notice the used of PREVIOUS to get the output from the Data Extraction task.
            previous_index = int(i/len(impls))
            tigres.log_debug("impl_task_previous_index","Col:{}, Impl:{}, PreviousIndex:{}".format(col,impl.__name__ ,previous_index))
            inputs_statistics.append(tigres.InputValues('Column', [tigres.PREVIOUS.i[previous_index]]))
            task_name = "Task.%s.%s" % (col, impl.__name__)
            tasks_statistics.append( tigres.Task(task_name, task_type=tigres.FUNCTION, impl_name=impl))
            i += 1


    # Invoke the  templates
    output = None
    try:
        tigres.split('Data Count', task_data_extraction, inputs_data_extraction,
                              tigres.TaskArray(tasks=tasks_col_data), tigres.InputArray(values=input_data))
        # Print out the output returns by the split template
        print("Columns being analyzed")
        for col in tigres.get_results():
            print("\t{}".format(col[1]))

        tigres.parallel('Data Count Categories', tigres.TaskArray(tasks=tasks_col_categories),
                        tigres.InputArray(values=input_col_data))


        print("Column Count: {}, Statistics Tasks:{}".format(len(tigres.get_results()), len(tasks_statistics)))
        for values, category in tigres.get_results():
            print("{} - Most Cases:".format(category))
            for v in collections.Counter(values).most_common(5):
                print("\t{}".format(v))

        tigres.parallel('Data Stats on Categories', tigres.TaskArray(tasks=tasks_statistics),
                        tigres.InputArray(values=inputs_statistics))

        output = tigres.get_results()
        output.sort()
        for o in output:
            print(o)

    except tigres.utils.TigresException as e:
        print(e)

        # Check the status of all tasks in the log
        tasks_to_check = [task.name for task in tasks_col_categories]
        for task in tasks_col_data: tasks_to_check.append(task.name)
        for task in tasks_statistics: tasks_to_check.append(task.name)
        task_check = tigres.check('task', names=tasks_to_check, state=tigres.utils.State.FAIL)

        if task_check:
            for task_record in task_check:
                print((".. State of {} = {} -  {}".format(task_record.name, task_record.state, task_record.errmsg)))
        exit(-1)

    tigres.dot_execution()


if __name__ == "__main__":

    if len(sys.argv) <= 2:
        print("\nBasic Counts Workflow \n")
        print(("Usage: {} <execution plugin> <data file> <columns>\n".format(
            sys.argv[0])))
        print("\texecution plugin - possible values are ({})".format(
            "|".join(list(tigres.utils.Execution.LOOKUP.keys()))))
        print("\tdata file - comma delimited text file with a header row")
        print("\tcolumns (Optional) list of column number (1,2,3) or blank for all")
        exit()
    import timeit

    execution = tigres.utils.Execution.get(sys.argv[1])
    print(
        "{} seconds".format(timeit.timeit(stmt='import tigres; main("{}", sys.argv[2], sys.argv[3])'.format(execution),
                                          setup="from __main__ import main".format(__name__),
                                          number=1)))

