#!/usr/bin/env python
"""
Tigres Split Template example.

This example takes a data files, extracts the data
and performs basic statistics on some of the data (average, median, total) using
a Tigres split template.
"""

import os
import sys
import math

import tigres
import statistics


def extract_data(filename, has_header=True):
    """
    Get the data from the csv file
    :param filename:
    :return: list of lists
    :rtype: list
    """
    column_headers = None
    data = []
    with open(filename, 'rU') as myfile:
        for line in myfile:
            values = line.strip("\n").split(",")
            if not column_headers and has_header:
                column_headers = values
            else:
                data.append(values)

    return data, column_headers



def median(numeric_values):
    """
    Calculate the median of the given numeric values

    :param numeric_values:
    :type numeric_values: list
    :return: median
    :rtype: float
    """
    values = sorted(numeric_values)
    if len(values) % 2 == 1:
        return values[int((len(values) + 1) / 2 - 1)]
    else:
        lower = values[int(len(values) / 2 - 1)]
        upper = values[int(len(values) / 2)]
        return (float(lower + upper)) / 2


def file_col_generator(data, col):
    """
    Generator for the values in the data object for the given column
    :param data:
    :type data: list
    :param col:
    :type col: int
    """
    for values in data:
        value = float(values[col])
        yield value


def average_by_column(data, column_num, column_names):
    """
    Calculate the average of the given numeric values in the specified column

    :param data:
    :type data: list
    :param column_num:
    :type column_num: int
    :param column_names: the list of column names
    :type column_names: list
    :return: (operation, column_name, average)
    :rtype: tuple
    """
    values = []
    for value in file_col_generator(data, column_num):
        values.append(value)
    total = sum(values)
    count = len(values)
    average = total / count
    return average_by_column.__name__, column_names[column_num], average


def total_by_column(data, column_num, column_names):
    """
    Calculate the total of the given numeric values in the specified column

    :param data:
    :type data: list
    :param column_num:
    :type column_num: int
    :param column_names: the list of column names
    :type column_names: list
    :return: (operation, column_name, total)
    :rtype: tuple
    """
    values = []
    for value in file_col_generator(data, column_num):
        values.append(value)
    total = sum(values)
    return total_by_column.__name__, column_names[column_num], total


def median_by_column(data, column_num, column_names):
    """
    Calculate the median of the given numeric values in the specified column

    :param data:
    :type data: list
    :param column_num:
    :type column_num: int
    :param column_names: the list of column names
    :type column_names: list
    :return: (operation, column_name, median)
    :rtype: tuple
    """
    values = []
    for value in file_col_generator(data, column_num):
        values.append(value)

    return median_by_column.__name__, column_names[column_num], median(values)


def stdev_by_column(data, column_num, column_names):

    values = []
    for value in file_col_generator(data, column_num):
        values.append(value)

    return stdev_by_column.__name__, column_names[column_num], math.sqrt(statistics.variance(values))


def variance_by_column(data, column_num, column_names):

    values = []
    for value in file_col_generator(data, column_num):
        values.append(value)

    return variance_by_column.__name__, column_names[column_num], statistics.variance(values)


def main(execution, data_file, columns):
    # Initialize the Tigres program
    tigres.start(name="Basic Statistics By Column", log_dest=os.path.splitext(__file__)[0] + '.log', execution=execution)

    tigres.set_log_level(tigres.Level.DEBUG)

    # Initialize the field values that we are interested in
    input_cols = [int(c) for c in columns.split(',')]

    # Initialize the operations we would like to perform on the data columns
    impls = [average_by_column, total_by_column, median_by_column, stdev_by_column, variance_by_column]

    # Data Extraction Task
    task_data_extraction = tigres.Task("Data Extraction", tigres.FUNCTION, impl_name=extract_data, input_types=[str])
    inputs_data_extraction = tigres.InputValues("Data File", [data_file])


    # Define Tasks for statistics
    tasks_statistics = []
    inputs_statistics = []
    for col in input_cols:
        for impl in impls:
            # Notice the used of PREVIOUS to get the output from the Data Extraction task.
            inputs_statistics.append(tigres.InputValues('Column', [tigres.PREVIOUS.task_data_extraction.i[0],
                                                            col, tigres.PREVIOUS.task_data_extraction.i[1]]))
            task_name = "Task.%s.%s" % (col, impl.__name__)
            tasks_statistics.append(
                tigres.Task(task_name, task_type=tigres.FUNCTION, impl_name=impl))


    # Invoke the split template
    output = None
    try:
        output = tigres.split('Data Stats', task_data_extraction, inputs_data_extraction,
                       tigres.TaskArray(tasks=tasks_statistics), tigres.InputArray(values=inputs_statistics))
    except tigres.utils.TigresException as e:
        print(e)

        # Check the status of the statistical tasks in the log
        task_check = tigres.check('task', names=[task.name for task in tasks_statistics], state=tigres.utils.State.FAIL)
        if task_check:
            for task_record in task_check:
                print(".. State of {} = {} -  {}".format(task_record.name, task_record.state, task_record.errmsg))
        exit(-1)

    # Print out the output returns by the split template
    tigres.log_info("result_count",message="Parallel Output Result Count {}".format(len(output)))
    for output in output:
        if isinstance(output, tigres.utils.TaskFailure):
            print(output.error)
        else:
            tigres.log_info("results", message=output)

    tigres.dot_execution()



if __name__ == "__main__":

    try:
        import timeit
        execution = tigres.utils.Execution.get(sys.argv[1])
        time = timeit.timeit(stmt='import tigres; main("{}", sys.argv[2], sys.argv[3])'.format(execution),
              setup="from __main__ import main".format(__name__),
              number=1)
        tigres.log_info("performance", message="time={}, execution={}, file={}, cols={}".format(time,sys.argv[1],sys.argv[2], sys.argv[3]))
        print ("time={}, execution={}, file={}, cols={}".format(time,sys.argv[1],sys.argv[2], sys.argv[3]))
    except tigres.utils.TigresException as e:
        print(e)

        # Check the status of the statistical tasks in the log
        task_check = tigres.check('task', state=tigres.utils.State.FAIL)
        if task_check:
            for task_record in task_check:
                print(".. State of {} = {} -  {}".format(task_record.name, task_record.state, task_record.errmsg))

        log_records = tigres.query(["event = extract_data"])
        for record in log_records:
            print("{} - {}".format(record.event, record.message))

