#!/usr/bin/env python
"""
This workflow demonstrates launching parallel tasks that have data dependencies.

"""
from __future__ import print_function

import os
import sys

import tigres


def check_status():
    """
    Are there any failed tasks?
    """
    task_check = tigres.check('task', state=tigres.utils.State.FAIL)

    if not task_check:
        results = tigres.get_results()
        print(results)
    else:
        raise tigres.utils.TigresException("A Task Failure was found")


def add_up():
    import fileinput
    total = 0
    count = 0
    for line in fileinput.input():
        total += float(line)
        count += 1
    print("sum:{}".format(total), end=" ")
    return "count:{}".format(count)


def main(execution):
    tigres.start(name="Example Workflow Sequence Pipeline", log_dest='example_workflow_sequence_pipeline.log',
                 execution=execution)
    failed = 0
    method = "pipe"
    try:

        # Create the Tasks
        task_grep = tigres.Task("grep", tigres.EXECUTABLE, "grep")
        task_awk = tigres.Task("awk", tigres.EXECUTABLE, "awk")
        task_sort = tigres.Task("sort", tigres.EXECUTABLE, "sort")
        task_unique = tigres.Task("uniq", tigres.EXECUTABLE, "uniq")
        task_add_up = tigres.Task("add_up", tigres.FUNCTION, add_up)

        # The absolute of the file/pipe
        inputs_add_up = tigres.InputValues("inputs_add_up_{}".format(method),
                                           [])

        inputs_grep = tigres.InputValues("inputs_grep_{}".format(method), ["write", "{}/data.csv".format(os.getcwd())])

        inputs_awk = tigres.InputValues("inputs_{}_awk".format(method),
                                        ["-F,", "'{{print $2}}'".format(os.getcwd())])

        inputs_sort = tigres.InputValues("inputs_{}_sort".format(method),
                                         [])

        inputs_uniq = tigres.InputValues("inputs_{}_uniq".format(method),
                                         [])

        output = tigres.sequence("Pipeline Sequence (executable only)", tigres.TaskArray(name="Pipeline Executables",
                                                                                         tasks=[task_grep, task_awk,
                                                                                                task_sort,
                                                                                                task_unique]),
                                 tigres.InputArray(values=[inputs_grep, inputs_awk, inputs_sort, inputs_uniq]),
                                 redirect=True)

        print("\nPipeline Template (execution only)")
        check_status()
        assert (output == "server-6204\nserver-6401")

        print("\nPipeline run Commands (mixed)")
        inputs_awk = tigres.InputValues("inputs_{}_awk".format(method),
                                        ["-F,", "'{{print $5}}'".format(os.getcwd())])
        output = tigres.sequence("Pipeline Sequence Mixed", tigres.TaskArray(name="Pipeline Executables",
                                                                             tasks=[task_grep, task_awk, task_add_up]),
                                 tigres.InputArray(values=[inputs_grep, inputs_awk, inputs_add_up]),
                                 redirect=True)
        check_status()
        assert (output == "sum:3360.0 count:85")

    except tigres.utils.TigresException as e:
        failed = 1
        print(e)

    task_check = tigres.check('task', state=tigres.utils.State.FAIL)
    if task_check:
        # clean up files
        failed = 1
        # There was a failure. Print the tasks that failed
        print("Failed!!", file=sys.stderr)
        for task_record in task_check:
            print(".. Task Failed: {}, {}".format(task_record.name, task_record.errmsg), file=sys.stderr)
    elif not failed:
        print("Success!!")

    # Print the Graph
    # tigres.dot_execution()

    # ################################
    # End the Workflow
    # ################################
    tigres.end()

    if failed:
        raise
    return failed


if __name__ == "__main__":

    if len(sys.argv) <= 1:
        print(("Usage: {} ({})>".format(sys.argv[0], "|".join(list(tigres.utils.Execution.LOOKUP.keys())))))
        exit()
    main(tigres.utils.Execution.get(sys.argv[1]))
