import glob
import os
import shutil
import signal
import subprocess
import threading
import time
from tigres import end
from tigres.utils import Execution
from tigres.core.execution.plugin.distribute  import (
                            ExecutionPluginDistributeProcess)
from tests.unit.execution.test_elastic import write_env_file

__author__ = 'gonzalo'

import unittest

def get_tigres_clients_count(server_key):
    """Returns number of processes which include server_key in their command."""
    
    ps_command='/bin/ps aux'
    sub_proc = subprocess.Popen(ps_command.split(), stdout=subprocess.PIPE)
    count=0
    output, err = sub_proc.communicate()

    for line in output.decode("utf-8").split("\n"):
        if server_key in line:
            count+=1
    sub_proc.stdout.close()
    return count

def signal_sender_action(test_obj, use_file=False, only_one_host=False):
    """Simulates the action of eHPC-DMCTP when the application receives more
    resouces to executie. It re-writes the TIGRES_HOSTS environment variable
    and sending a signal to the main Tigres process. Runs tests to detect if
    the application scaled accordingly. 
    :param test_obj: unittest.TestCase object used to call asserts.
    :param use_file: boolean if False, the new number of avaiable hosts will
        be set in the TIGRES_HOST env var. If True, TIGRES_HOSTS will be set
        in a file that will be read by Tigres to re-set its environment
        variables.
    :param only_one_host: if False, it emulates the scale up from one to two
        worker nodes. If True, it emulates a scale from one to one worker node.
    """
    print("Signal sender sleeps for 3 seconds.")    
    time.sleep(3)
    print ("Signal is sent to the application, it should be ignored since it is "
           "executing a sequence pattern")
    os.kill(os.getpid(), signal.SIGUSR1)
    
    print("Signal sender sleeps for 4 seconds. ")    
    time.sleep(4)
    print("Signal sender wakes up and sends a signal to the main process")
    test_obj.assertEqual(len(ExecutionPluginDistributeProcess.programs), 1)
    new_host_string = "localhost,127.0.0.1"
    if only_one_host:
        new_host_string="localhost"
    if not use_file:
        os.environ["TIGRES_HOSTS"]=new_host_string
    else:
        file_route=write_env_file(test_obj,
                                  {"TIGRES_HOSTS":new_host_string})
        os.environ["TIGRES_ENV_FILE"] = file_route
    secret_key = ExecutionPluginDistributeProcess.secret_key
    proc_count_before=get_tigres_clients_count(secret_key)
    os.kill(os.getpid(), signal.SIGUSR1)
    time.sleep(1)
    proc_count_after=get_tigres_clients_count(secret_key)
    try:
        if only_one_host:
            test_obj.assertEqual(len(ExecutionPluginDistributeProcess.programs),
                                 1)
            test_obj.assertEqual(proc_count_after, proc_count_before)
        else:
            test_obj.assertEqual(len(ExecutionPluginDistributeProcess.programs),
                                 2)
            test_obj.assertGreater(proc_count_after, proc_count_before)
    except AssertionError as e:
        test_obj._bad_ending=True
        test_obj._exception=e
    
def signal_handler_fail(self,arg1, arg2):
    """This signal handler is pre-programmed to checkt that the elastic module
    does not let signals arrive to the application."""
    raise Exception("Signal was not captured by module.")
    
class TestScaleUp(unittest.TestCase):
    def setUp(self):
        self.save_env_vars(["TIGRES_HOSTS", "OTIGRES_PATH", 
                            "OTIGRES_LD_LIBRARY_PATH", "TIGRES_ENV_FILE"])
        self._cwd = os.path.split(__file__)[0]
        if not self._cwd:
            self._cwd = "."
        self._cwd_output = self._cwd + "/output"

        if os.path.exists(self._cwd_output):
            shutil.rmtree(self._cwd_output)
        os.mkdir(self._cwd_output)

        for f in ['task1impl', 'task2impl', 'data.csv']:
            shutil.copy(self._cwd + "/{}".format(f),
                        self._cwd_output + "/{}".format(f))
        os.chdir(self._cwd_output)

        self.prev_action = signal.signal(signal.SIGUSR1,
                                         signal_handler_fail)
        self._bad_ending=False
        self._exception=None
        #Configure environment to run clients
        env_root = os.getenv("VIRTUAL_ENV", None)
        if env_root is not None:
            if os.getenv("OTIGRES_PATH") is None:
                os.environ["OTIGRES_PATH"]=os.path.join(env_root,"bin")
            if os.getenv("OTIGRES_LD_LIBRARY_PATH") is None:
                os.environ["OTIGRES_LD_LIBRARY_PATH"]=os.path.join(env_root,
                                                                   "lib")
        if os.getenv("TIGRES_HOSTS") is None:
            os.environ["TIGRES_HOSTS"] = "localhost"
        if "TIGRES_ENV_FILE" in os.environ:
            del os.environ["TIGRES_ENV_FILE"]
    
    def tearDown(self):
        os.chdir(self._cwd)
        if os.path.exists(self._cwd_output):
            shutil.rmtree(self._cwd_output)
        for logfile in glob.glob("*.log"):
            os.unlink(logfile)
        self.restore_env_vars(["TIGRES_HOSTS", "OTIGRES_PATH", 
                            "OTIGRES_LD_LIBRARY_PATH", 
                            "TIGRES_ENV_FILE"])
    
    def save_env_vars(self, var_names):
        """Stores the state of a set of environment vars.
        :param var_name list of strings with the name of enviroment vars to be
            saved.
        """
        self._env_store={}
        for name in var_names:
            value = os.getenv(name, None)
            if value is not None:
                self._env_store[name]=value
    def restore_env_vars(self, var_names):
        """Restores the state of a list of environment vars that have which
        state was previously stored with save_env_vars. If an enviroment
        variable was not set before, it will be unset.
        :params var_names list of strings with the name of enviroment vars to be
            restored.
        """
        for name in var_names:
            value=os.getenv(name, None)
            if value is not None:
                if name in self._env_store.keys():
                    os.environ[name]=self._env_store[name]
                else:
                    del os.environ[name]            
        
    

    def test_elastic(self):
        """
        Runs a example timed with sleeps, changes TIGRES_HOSTS and sends a
        signal to the main process to trigger the scale up mechanism. 
        :return:
        """
        from tests.examples import example_sleeps
        for module in [example_sleeps]:

            if not os.name == 'nt':
                print ("\n####################################\n"
                       "DISTRIBUTE - Elastic.\n"
                       "####################################\n")
                # FIXME: This doesn't state on windows
                self.signal_sender()
                module.main(Execution.DISTRIBUTE_PROCESS)
                end()
        self.assertFalse(self._bad_ending, str(self._exception))
    def test_elastic_file(self):
        """
        Runs a example timed with sleeps, changes TIGRES_HOSTS and sends a
        signal to the main process to trigger the scale up mechanism. 
        :return:
        """
        from tests.examples import example_sleeps
        for module in [example_sleeps]:

            if not os.name == 'nt':
                print ("\n####################################\n"
                       "DISTRIBUTE - Elastic env vars in file.\n"
                       "####################################\n")
                # FIXME: This doesn't state on windows
                self.signal_sender(True)
                module.main(Execution.DISTRIBUTE_PROCESS)
                end()
        self.assertFalse(self._bad_ending, str(self._exception))
    def test_elastic_file_one(self):
        """
        Runs a example timed with sleeps, changes TIGRES_HOSTS and sends a
        signal to the main process to trigger the scale up mechanism. 
        :return:
        """
        from tests.examples import example_sleeps
        for module in [example_sleeps]:

            if not os.name == 'nt':
                print ("\n####################################\n"
                       "DISTRIBUTE - Elastic, do not scale.\n"
                       "####################################\n")
                # FIXME: This doesn't state on windows
                self.signal_sender(True, True)
                module.main(Execution.DISTRIBUTE_PROCESS)
                end()
        self.assertFalse(self._bad_ending, str(self._exception))
    def signal_sender(self, use_file=False, only_one=False):
        """
        Creates the thread to that will emulate the behavior of DMTCP when
        it re-starts a checkpointed Tigres workflow.
        :param use_file: boolean if False, the new number of avaiable hosts will
            be set in the TIGRES_HOST env var. If True, TIGRES_HOSTS will be set
            in a file that will be read by Tigres to re-set its environment
            variables.
        :param only_one_host: if False, it emulates the scale up from one to two
            worker nodes. If True, it emulates a scale from one to one worker
            node.
        """
        self._t = threading.Thread(target=signal_sender_action, 
                                   args=(self, use_file,only_one,))
        self._t.start()
        

if __name__ == '__main__':
    unittest.main()
