"""
2/23/16 by James Fox (jsfox@lbl.gov)
"""

import glob
import os
from tigres.core.state.program import Program
from tigres.utils import State
from tigres import check
import shutil
import unittest

from tests.integration.recovery.resources.recover_sequence import LOG_DEST, run_broken_sequence, run_fixed_sequence, run_fixed_sequence_m
from tests.integration.recovery.resources.recover_parallel import LOG_DEST_PARALLEL, run_broken_parallel, run_fixed_parallel
from tests.integration.recovery.resources.recover_split_merge import LOG_DEST_SM, run_broken_split_parallel, run_fixed_split_parallel \
    ,run_broken_merge_parallel, run_fixed_merge_parallel, run_broken_merge, run_fixed_merge, run_broken_split, run_fixed_split \
    ,run_fixed_merge_parallel_m

class Test(unittest.TestCase):

    @classmethod
    def setUpClass(cls):
        Program().clear()

        for logfile in glob.glob(os.path.join(os.path.dirname(os.path.abspath(__file__)), "resources/*.log")):
            os.unlink(logfile)
        for logfile in glob.glob(os.path.join(os.path.dirname(os.path.abspath(__file__)), "resources/*.dot")):
            os.unlink(logfile)
        for dirPath in glob.glob(os.path.join(os.path.dirname(os.path.abspath(__file__)), "resources/*data-logs")):
            shutil.rmtree(dirPath)

    @classmethod
    def tearDownClass(cls):
        for logfile in glob.glob(os.path.join(os.path.dirname(os.path.abspath(__file__)), "resources/*.log")):
            os.unlink(logfile)
        for logfile in glob.glob(os.path.join(os.path.dirname(os.path.abspath(__file__)), "resources/*.dot")):
            os.unlink(logfile)
        for dirPath in glob.glob(os.path.join(os.path.dirname(os.path.abspath(__file__)), "resources/*data-logs")):
            shutil.rmtree(dirPath)
        Program().clear()

    def test_recover_sequence1(self):
        print("testing sequence recovery, when there's nothing to recover\n")
        run_broken_sequence(recover=True, recovery_log=True)
        task_recs = check('task', state=State.RUN)
        self.assertEqual(len(task_recs), 3)

    def test_recover_sequence2(self):
        print("testing successful sequence recovery\n")
        run_broken_sequence(recover=False, recovery_log=True)
        run_fixed_sequence(recover=True, recovery_log=True)
        recovered = check('task', state=State.RECOVERED)
        rerun = check('task', state=State.RERUN)
        self.assertEqual(len(recovered), 2)
        self.assertEqual(len(rerun), 1)

    def test_recover_sequence3(self):
        print("testing failed sequence recovery due to lack of input/output data from previous run.\ntesting successful subsequent re-run without recovery.\n")
        try:
            run_broken_sequence(recover=False, recovery_log=False)
            run_fixed_sequence(recover=True)
        except Exception as e:
            #self.assertEqual(str(e), 'Missing inputs: task broken task\'s inputs could not be recovered.')
            self.assertEqual(check('task', state=State.DONE), None)

        run_fixed_sequence(recover=False)
        run_recs = check('task', state=State.RUN)
        self.assertEqual(len(run_recs), 4)

    def test_recover_sequence4(self):
        print("testing recovery on a previously successful workflow.\nfinal sequence task output should exist and be accessible.\n")
        run_fixed_sequence(recover=False, recovery_log=True)
        run_fixed_sequence(recover=True, recovery_log=True)
        recovered = check('task', state=State.RECOVERED)
        self.assertEqual(len(recovered), 4)
        recovery = Program().recovery
        tLog = recovery.getTaskLog("normal task 3", recovery.curr_prgm_id)
        recovered_with_data = check('task', names=["^" + "normal task 3"], task_log=tLog)
        self.assertNotEqual(recovered_with_data[-1].output, None)

    def test_recover_parallel1(self):
        print("testing parallel recovery, when there's nothing to recover\n")
        run_broken_parallel(recover=True, recovery_log=True)
        run_recs = check('task', state=State.RUN)
        self.assertEqual(len(run_recs), 4)
        failed_recs = check('task', state=State.FAIL)
        self.assertEqual(len(failed_recs), 1)

    def test_recover_parallel2(self):
        print("testing successful parallel recovery\n")
        run_broken_parallel(recover=False, recovery_log=True)
        run_fixed_parallel(recover=True, recovery_log=True)
        recovered = check('task', state=State.RECOVERED)
        rerun = check('task', state=State.RERUN)
        self.assertEqual(len(recovered), 3)
        self.assertEqual(len(rerun), 1)

    def test_recover_parallel3(self):
        print("testing failed parallel recovery due to lack of input/output data from previous run.\ntesting successful subsequent re-run without recovery.\n")
        try:
            run_broken_parallel(recover=False, recovery_log=False)
            run_fixed_parallel(recover=True)
        except Exception as e:
            #self.assertEqual(str(e), 'Missing inputs: task broken task\'s inputs could not be recovered.')
            self.assertEqual(check('task', state=State.DONE), None)

        run_fixed_parallel(recover=False)
        done_recs = check('task', state=State.DONE)
        self.assertEqual(len(done_recs), 4)

    def test_recover_parallel4(self):
        print("testing recovery on a previously successful parallel workflow.\nfinal parallel template's output should exist and be accessible.\n")
        run_fixed_parallel(recover=False)
        run_fixed_parallel(recover=True)

        recovered = check('task', state=State.RECOVERED)
        self.assertEqual(len(recovered), 4)

        recovery = Program().recovery
        task_names = ['normal task 1', 'normal task 2', 'broken task', 'normal task 3']
        for tName in task_names:
            tLog = recovery.getTaskLog(tName, recovery.curr_prgm_id)
            recovered_with_data = check('task', names=["^" + tName], task_log=tLog)
            self.assertNotEqual(recovered_with_data[-1].output, None)

if __name__ == "__main__":
    unittest.main()