import os
import signal
import unittest
import time
from tigres.core.execution.elastic import (redo_env_vars_from_file,
                                           init_elastic, ElasticRMCommunicator)

__author__ = 'gonzalo'

def write_env_file(test_obj, env_var, file_route = "/tmp/my_env_file.txt"):
    test_obj.addCleanup(os.remove, file_route)
    f = open(file_route, "w")
    for (key, value) in env_var.items():
        f.write("{0}={1}\n".format(key, value))
    f.close()
    return file_route

class TestElasticModuleFunctions(unittest.TestCase):
    def setUp(self):
        if os.getenv("MYVAR1"):
            del os.environ["MYVAR1"]
        if os.getenv("MYVAR2"):
            del os.environ["MYVAR2"]
        if os.getenv("TIGRES_ENV_FILE"):
            del os.environ["TIGRES_ENV_FILE"]

    def test_redo_env_vars_from_file(self):
        os.environ["MYVAR1"]="PREVALUE1"
        file_route= write_env_file(self, {"MYVAR1":"POSTVALUE1",
                                        "MYVAR2":"POSTVALUE2"})
        self.assertEqual(os.getenv("MYVAR1", None), "PREVALUE1")
        self.assertEqual(os.getenv("MYVAR2", None), None)
        os.environ["TIGRES_ENV_FILE"]=file_route
        self.assertTrue(redo_env_vars_from_file())
        self.assertEqual(os.getenv("MYVAR1", None), "POSTVALUE1")
        self.assertEqual(os.getenv("MYVAR2", None), "POSTVALUE2")
        
    def test_redo_env_vars_from_file_env_var_name(self):
        os.environ["MYVAR1"]="PREVALUE1"
        file_route= write_env_file(self, {"MYVAR1":"POSTVALUE1",
                                        "MYVAR2":"POSTVALUE2"}, )
        self.assertEqual(os.getenv("MYVAR1", None), "PREVALUE1")
        self.assertEqual(os.getenv("MYVAR2", None), None)
        os.environ["ENV_NAME"]=file_route
        redo_env_vars_from_file("ENV_NAME")
        self.assertEqual(os.getenv("MYVAR1", None), "POSTVALUE1")
        self.assertEqual(os.getenv("MYVAR2", None), "POSTVALUE2")
        
    def test_redo_env_vars_from_file_not_set(self):
        self.assertFalse(redo_env_vars_from_file())
    
    def test_init_elastic(self):
        init_elastic(signal.SIGUSR2)
        os.kill(os.getpid(), signal.SIGUSR2)

class TestElasticRMCommunicator(unittest.TestCase): 
    def test_callback_from_env(self):
        co = CallbackObj()
        init_elastic(signal.SIGUSR2)
        ermc = ElasticRMCommunicator(host_list_var="MYVAR", no_env_file=True)
        ermc.set_scale_callback(co.callback)
        os.environ["MYVAR"]="h1,h2,h3"
        
        os.kill(os.getpid(), signal.SIGUSR2)
        time.sleep(1)
        self.assertEqual(co.called,0)
        
        ermc.enable()
        os.kill(os.getpid(), signal.SIGUSR2)
        time.sleep(1)
        self.assertEqual(co.called,1)
        self.assertEqual(co.hostlist,"h1,h2,h3")
    
    def test_callback_from_file(self):
        co = CallbackObj()
        init_elastic(signal.SIGUSR2)
        ermc = ElasticRMCommunicator(host_list_var="MYVARINFILE",
                                     no_env_file=False,
                                     env_file_route_var="ENVFILEROUTE")
        ermc.set_scale_callback(co.callback)
        os.environ["ENVFILEROUTE"]=write_env_file(self, {"MYVARINFILE":"h4,h5"})
        
        os.kill(os.getpid(), signal.SIGUSR2)
        time.sleep(1)
        self.assertEqual(co.called,0)
        
        ermc.enable()
        os.kill(os.getpid(), signal.SIGUSR2)
        self.assertEqual(co.called,1)
        self.assertEqual(co.hostlist,"h4,h5")
    
    def test_log_functions(self):
        fl = FakeLog()
        ermc = ElasticRMCommunicator()
        ermc.set_log(fl.log, FakeLog.level_info, FakeLog.level_error)
        
        ermc._log_info("info1")
        ermc._log_error("error1")
        ermc._log_info("info2")
        ermc._log_error("error3")
        self.assertEqual(fl._info_msgs, ["info1", "info2"])
        self.assertEqual(fl._error_msgs, ["error1", "error3"])
        
class FakeLog(object):
    level_info=1
    level_error=2
    def __init__(self):
        self._info_msgs=[]
        self._error_msgs=[]
    
    def log(self, level, msg):
        if level==FakeLog.level_info:
            self._info_msgs.append(msg)
        elif level==FakeLog.level_error:
            self._error_msgs.append(msg)
            
class CallbackObj(object):
    def __init__(self):
        self.called=0
        self.hostlist=[]
    def callback(self, hostlist):
        self.hostlist=hostlist
        self.called+=1
    
        
if __name__ == '__main__':
    unittest.main()