"""
`tigres.core.execution.elastic`
========================

.. currentmodule:: tigres.core.execution.elastic

:platform: Unix, Mac
:synopsis: enables communication with a resource manager (RM) that may
    increase or decrease the resource allocation assigned to a tigres workflow
    executing in distributed mode. If a new resorce allocation is offered
    to the workflow, the external RM should:
    1) Store the list of hostnames assigned to the workflow by either:
        a) re-setting TIGRES_HOSTS in the environment of the parent process (
        this code overrides the python environment sandbox and reads from the
        parent environment).
        b) setting the TIGRES_ENV_FILE environment var pointing to a file that
        contains the line 'TIGRES_HOSTS=host1,host2,hostX". This option
        overrides and setting of TIGRES_HOSTS.
    b) Notify the tigres workflow that a new list of hostnames is available
    by sending a SIGUSR1 signal to the process running the tigres master.
    
    By default this module is enabled in Tigres.
Classes
=======
 * :py:class:`tigres.core.execution.elastic.ElasticRMCommunicator` - Comms
     class to be instanciated by the execution plug-in.

.. moduleauthor:: Gonzalo Rodrigo <gprodrigoalcarez@lbl.gov>

"""
import os
import signal
import subprocess


def init_elastic(elastic_signal=signal.SIGUSR1):
    """
    Sets the signal chosen to notify a change in resource allocation. 
    and configuring an "IGNORE" behavior for it.
    
    :param elastic_signal: signal type (from signal python module) that
        represents the operative system signal that RM will send to tigres
        master to notigy a change in resources.
    """ 
    ElasticRMCommunicator._dmtpc_signal=elastic_signal
    
    signal.signal(elastic_signal, signal.SIG_IGN)
    

class ElasticRMCommunicator(object):
    """
    Communications class that can receive notifications of resource changes
    from external RM to call a programmable callback passing the list of new
    hostnames.
    """
    _dmtpc_signal=signal.SIGUSR1
    
    def __init__(self, host_list_var="TIGRES_HOSTS", no_env_file=False,
                 env_file_route_var="TIGRES_ENV_FILE"):
        """Initilizes the RM communicator class.
        :param host_list_var: sting with the environment variable name
            containing the list of hostnames separated by commas that the
            tigres uses in distributed mode.
        :param no_env_file: If set to True, the class will not read
            host_list_var from a file.
        :param env_file_route_var: string with the environment variable name
            containing a filesystem route pointing to a text file. Each line
            of this text file will have the format
                [env_var_name]=[env_var_value]
            so the listed environment variables as set wihtin Tigres. This is
            used to communicate the new TIGRES_HOSTS through a file.
        """
        self._host_list_var = host_list_var
        self._no_env_file=no_env_file
        self._env_file_route_var=env_file_route_var
        
        self._scale_callback=None
        self._log_f=None
        self._level_info_id=None
        self._level_error_id=None
    
    def set_scale_callback(self, callback=None):
        """
        Sets the function to be called when the _dmtcp_singal arrives
        to the process.
        
        :param callback: function with the signature 
            def callback(host_list)
            where host_list is a string containing comma separated hostnames
            of the nodes made available to Tigres.
        """
        self._scale_callback=callback
    
    def set_log(self, log_f, info_level, error_level):
        """
        Inits the log functions to be used by the class.
        :param log_f: function of with the signature:
            log(log_level, msg_to_log)
        :param info_level: log_level value used for INFO messages.
        :param error_level: log_level value used for error messages.
        """
        self._log_f=log_f
        self._level_info_id = info_level
        self._level_error_id = error_level
    
    def _log_info(self, msg):
        """Logs msg as an info log entry."""
        if self._level_info_id!=None and self._log_f!=None:
            self._log_f(self._level_info_id, msg)
    def _log_error(self, msg):
        """Logs msg as an error log entry."""
        if self._level_error_id!=None and self._log_f!=None:
            self._log_f(self._level_error_id, msg) 
    
    def enable(self):
        """Programs handling the OS signal configured in init_elastic to run
        the callback programmed by set_scale_callback."""
        self.prev_action = signal.signal(ElasticRMCommunicator._dmtpc_signal,
                                         self.signal_handler)
    def disable(self):
        """Programs handling the OS signal configured in init_elastic to
        ignore."""
        signal.signal(ElasticRMCommunicator._dmtpc_signal, signal.SIG_IGN)
    
    def signal_handler(self, arg1, arg2):
        """Handles a signal by running an external callback programmed by
        set_scale callback."""
        if (arg1 != ElasticRMCommunicator._dmtpc_signal):
            return 0
        self._log_info("{0} signal received, running reading env vars and "
                       "running callback".format(
                                        ElasticRMCommunicator._dmtpc_signal))
        good_read=False
        if not self._no_env_file:
            good_read=redo_env_vars_from_file(
                                env_var_route_name=self._env_file_route_var)
           
            self._log_info("{1} after read env values from file: {0}"
                           "".format(os.getenv(self._host_list_var),
                                     self._host_list_var))
        if not good_read:
                custom_val= get_deep_env(self._host_list_var,
                                     os.getenv('self._host_list_var',None))
                if custom_val is not None:
                    os.environ[self._host_list_var] = custom_val
        
        if self._scale_callback:
            self._scale_callback(os.getenv(self._host_list_var))
    
        
def get_deep_env(env_var, default_var=None):
    """
    Reads the value of an en var of the OS process running this python
    program (In opposition to the value that could be in os.environ, which is
    set at the program start).
    
    :param en_var: string with the name of the environment variable to read.
    :param default_var: string with the default value to return if the variable
        is not set-
    :return: string, the value of the environment variable or, default_var if not
        set.
    """
    command=["python", "-c", 
            "import os; print(os.getenv('{0}', '{1}'))".format(env_var,
                                                               default_var)]
    
    p=subprocess.Popen(command,
                     stderr=subprocess.PIPE, stdout=subprocess.PIPE)
    output, err = p.communicate()
    output=output.decode("utf-8").strip("\n")
    if output ==  "" or output == "None": 
        return default_var
    else:
        return output

def redo_env_vars_from_file(env_var_route_name="TIGRES_ENV_FILE"):
    """
    Reads a list of env var sets from a file and writes them on os.environ.
    :param env_file_route_var: string with the environment variable name
        containing a filesystem route pointing to a text file. Each line
        of this text file will have the format
            [env_var_name]=[env_var_value]
        and each os[env_var_name] will be set to its corresponding
        env_var_value.
    """
    file_route=os.getenv(env_var_route_name, None) 
    if file_route is not None:
        if not os.path.exists(file_route):
            raise ValueError("TIGRES_ENV_FILE"
                      " is set but env file ({0}) does not exist!!!"
                      "".format(file_route))
        else:
            env_file = open(file_route, "r")
            env_text=env_file.readlines()
            for line in env_text:
                line=line.strip()
                key=line.split("=")[0]
                value="=".join(line.split("=")[1:])
                os.environ[key]=value
            env_file.close()
            return True
    return False
    
