"""
`tigres.core.execution.plugin`
==============================

.. currentmodule:: tigres.core.execution.plugin

:platform: Unix, Mac
:synopsis: The core tigres execution plugin api

Classes
=======
 * :py:class:`ExecutionPluginBase` - Base class for Execution Plugins inherits from `types.ModuleType`.

.. moduleauthor:: Val Hendrix <vchendrix@lbl.gov>

"""
from tigres.core.utils import get_metaclass, TigresInternalException

__all__ = ['job', 'local', 'distribute']

from abc import abstractmethod, ABCMeta
import types


class ExecutionPluginBase(types.ModuleType,
                          get_metaclass(ABCMeta, 'ExecutionPluginBaseABCMeta')):
    """
    Base class for Execution Plugins inherits from `types.ModuleType`.
    """

    @classmethod
    @abstractmethod
    def execute_sequence(cls, work_array, execution_data):
        """
        Executes the Sequence tasks in a pipeline. No intermediate data is collected

        """
        raise NotImplementedError

    @classmethod
    @abstractmethod
    def execute_task(cls, work, execution_data):
        """
        Executes a task and returns the output

        """
        raise NotImplementedError

    @classmethod
    def execute(cls, work, execution_data):
        if work["work_type"] == "task":
            return cls.execute_task(work, execution_data)
        elif work["work_type"] == "sequence":
            return cls.execute_sequence(work, execution_data)
        else:
            raise TigresInternalException("Invalid work for {}: type {}".format(work["name"], work["work_type"]))


    @classmethod
    @abstractmethod
    def startup(cls, run_fn, work_list=None):
        """
        Initializes the Execution plugin for parallel execution.

        :param run_fn: function to execute work unit with, it should take a WorkUnit as input
        :param work_list: The parallel state object to execute.
        :type work_list: tigres.core.state.WorkParallel

        """
        raise NotImplementedError

    @classmethod
    @abstractmethod
    def shutdown(cls):
        """
        Waits for work to finish and then shuts down the plugin engine

        """
        raise NotImplementedError


    @classmethod
    def add_work(cls, work):
        """
        add work to be executed

        :param work: adds work unit to plugin engine for execution
        """
        raise NotImplementedError
