"""
`tigres.core.execution.plugin.distribute`
========================

.. currentmodule:: tigres.core.execution.plugin.distribute

:platform: Unix, Mac
:synopsis: Execution Distributed plugin API

Classes
=======
 * :py:class:`ExecutionPluginDistributeProcess` - Execution plugin for distributing processes.



.. moduleauthor:: Val Hendrix <vchendrix@lbl.gov>

"""
import os
import socket
import subprocess
import uuid

from tigres.core.utils import get_free_port
from tigres.core.execution.plugin.local import ExecutionPluginLocalBase
from tigres.core.execution.utils import TaskServer, TaskClient
from tigres.core.execution.elastic import ElasticRMCommunicator


from tigres.core.monitoring.log import log_node
from tigres.core.monitoring.common import Level, NodeType



class ExecutionPluginDistributeProcess(ExecutionPluginLocalBase):
    """
    Execution plugin for distributing processes.
    """

    # I hate doing this class variables but it is by far the quickest
    # way for this *prototype* code. It is okay as long as the
    # methods used in the core code are called
    # correctly. When there is time, this should be refactored.
    task_server = None
    programs    = None
    task_client_process = None
    running_hosts = None
    prev_action = None 
    secret_key = None
    elasticrm_receiver = None

    @classmethod
    def startup(cls, run_fn, work_list=None):
        """ Executes in parallel across multiple hosts.
        Shared by parallel, split and merge templates

        Distributes tasks across the specified hosts machines. A
        `TaskServer` is run in the Tigres program.  `TigresClient`s which run workers
        that consume the tasks from the `TaskServer` queue are run on the
        specified hosts. If no host machines are specified the `TaskServer` and `TaskClients`
         will be run on the local machine.
         
        Enables elastic behavior: new resources will be used if allocation is
        scaled up.

        Environment variables:

          * `TIGRES_HOSTS` - a comma separated list of hosts names for distributing tasks.
                            (i. e. export TIGRES_HOSTS=host1,host2,host3)
          * `OTIGRES_*` - Any environment variable prefixed with `OTIGRES_` will pass to the
                          `tigres-client` program. (i.e export OTIGRES_PATH=tigres_home/bin)
                          will be translated to the command PATH=tigres_home/bin; tigres-client...)

        :param run_fn: function to state state with, it should take a WorkUnit as input
        :param work_list: The parallel state object to execute.
        :type work_list: tigres.core.state.WorkParallel

        """
        cls.configure_elastic(True)
        cls.startup_internal(run_fn, work_list=work_list)
        
    @classmethod
    def claim_new_workers(cls, host_list):
        """
        Sets host_list as the current TIGRES_HOSTS and create workers in the
        new nodes listed in it.
        :param host_list: a string containing comma separated hostnames
            of the nodes made available to Tigres.
        """
        os.environ["TIGRES_HOSTS"]=host_list
        cls.startup_internal(run_fn=cls.current_run_fn, 
                             work_list=cls.current_work_list,
                             already_running=True)
        
    @classmethod
    def startup_internal(cls, run_fn, work_list=None, already_running=False):
        """
        Actual actions for start_up.
        :param run_fn: function to state state with, it should take a WorkUnit
            as input
        :param work_list: The parallel state object to execute.
        :param already_running: If True, it creates as many workers as (number
            of hosts in TIGRES_HOSTS)-(currently running workers) 
        :type work_list: tigres.core.state.WorkParallel
        """
        ExecutionPluginDistributeProcess._log(Level.INFO,"STARTUP: {0}".format(
                                                            already_running))
        if not already_running:
            secret_key = str(uuid.uuid4())
            port = get_free_port()
            task_server = TaskServer(host=socket.gethostname(),
                                 port=port, secret_key=secret_key, work_list=work_list)
        else:
            secret_key = ExecutionPluginDistributeProcess.secret_key
            task_server=ExecutionPluginDistributeProcess.task_server 
            port=task_server._address[1]
        ExecutionPluginDistributeProcess.task_server = task_server
        ExecutionPluginDistributeProcess.current_run_fn = run_fn
        ExecutionPluginDistributeProcess.current_work_list = work_list
        ExecutionPluginDistributeProcess.secret_key = secret_key

        if ExecutionPluginDistributeProcess.running_hosts is None:
            ExecutionPluginDistributeProcess.running_hosts = []
        # Launch Client(s)
        hosts = None
        if 'TIGRES_HOSTS' in os.environ:
            hosts = os.environ['TIGRES_HOSTS']
        elif already_running:
            ExecutionPluginDistributeProcess._log(Level.INFO, 
                "Startup re-run not allowed for local run")
            return
        if not hosts:
            ExecutionPluginDistributeProcess._log(Level.INFO, "Using only localhost")

            def worker():
                task_client = TaskClient(cls.execute, host=socket.gethostname(),
                                         port=port, secret_key=secret_key)
                task_client.run()

            import multiprocessing
            task_client_process = multiprocessing.Process(target=worker)
            task_client_process.start()
            if not task_client_process:
                task_server.kill()
                raise Exception("There were no Tigres Clients Started")
            ExecutionPluginDistributeProcess.task_client_process = task_client_process
        else:
            host_list = hosts.split(',')
            if not already_running:
                programs = []
            else:
                programs = ExecutionPluginDistributeProcess.programs
                new_host_list = [host for host in host_list if not host in
                                 ExecutionPluginDistributeProcess.running_hosts]
                ExecutionPluginDistributeProcess._log(Level.INFO,
                    "Startup running during phase execution old nodes ({0}) vs."
                    "\n   new ({1}) and"
                    "\n   all ({2})"
                    "\n   TIGRES_HOSTS".format(
                        ExecutionPluginDistributeProcess.running_hosts,
                        new_host_list,
                        host_list))
                
            ExecutionPluginDistributeProcess._log(Level.INFO, "Using host_list %s" % str(host_list))
            env =  " ".join(
                ["export {}={};".format(k.replace('OTIGRES_', ''),
                                        os.environ[k]) for k in os.environ if
                 k.startswith('OTIGRES_')])
            command = "bash --login -c '{} tigres-client {} {} {}'".format(env,
                                                                               socket.gethostname(),
                                                                           port,
                                                                           secret_key)
            for host in host_list:
                if already_running and not host in new_host_list:
                    continue
                program=None
                try:
                    ExecutionPluginDistributeProcess._log(Level.INFO, "Calling ssh: %s" % host)
                    program = subprocess.Popen(
                        ["ssh", "-o", "StrictHostKeyChecking=no", host,
                         command], stderr=subprocess.PIPE, stdout=subprocess.PIPE)
                    ExecutionPluginDistributeProcess._log(Level.INFO, "Called ssh: %s" % host)     
                    if not program.returncode:
                        programs.append(program)
                    else:
                        if program.poll():
                            program.terminate()
                except Exception as err:
                    program.terminate()
                    programs.remove(program)
                    ExecutionPluginDistributeProcess._log(Level.ERROR, "ssh error %s" % str(err))

            if not len(programs) > 0:
                task_server.kill()
                raise Exception("There were no Tigres Clients Started")

            ExecutionPluginDistributeProcess.programs = programs
            ExecutionPluginDistributeProcess.running_hosts=host_list

    @classmethod
    def shutdown(cls):
        """
        Shutdown the engine
        """
        # Wait for all the jobs to finish
        if ExecutionPluginDistributeProcess.task_server:
            ExecutionPluginDistributeProcess.task_server.join()
            ExecutionPluginDistributeProcess.task_server = None

        # Close down the TaskClient(s)
        if ExecutionPluginDistributeProcess.task_client_process is not None:
            if ExecutionPluginDistributeProcess.task_client_process.is_alive():
                ExecutionPluginDistributeProcess.task_client_process.terminate()

        # Close down distributed processes
        if ExecutionPluginDistributeProcess.programs is not None:
            for p in ExecutionPluginDistributeProcess.programs:
                if p.poll():
                    p.terminate()
        ExecutionPluginDistributeProcess.running_hosts = []
        cls.configure_elastic(False)

    @classmethod
    def add_work(cls, work):
        """
        add work to be executed
        """

        ExecutionPluginDistributeProcess.task_server.submit(work)
    
    @staticmethod
    def _log(level, event, nodetype=NodeType.PROGRAM, **kwargs):
        """
        Log messages
        :param level:
        :param status:
        :param kwargs:
        :return:
        """
        log_node(level, "Distribute Process", event=event, nodetype=nodetype,
                 **kwargs)

    @classmethod
    def configure_elastic(cls, enable=True):
        """
        Initializes the ElasticRMCommunicator object and enables/disables
        its response to the external communications.
        
        :param enable: if True, the external RM signals will be interpreted
            and the local claim_new_workers function will be called (to create
            new workers if possible). If False, any signal of new resources
            will be ignored.
        """
        if cls.elasticrm_receiver is None:
            cls.elasticrm_receiver = ElasticRMCommunicator()
            cls.elasticrm_receiver.enable()
            cls.elasticrm_receiver.set_scale_callback(cls.claim_new_workers)
            cls.elasticrm_receiver.set_log(
                                        ExecutionPluginDistributeProcess._log,
                                        Level.INFO, Level.ERROR)
        
        if enable:
            cls.elasticrm_receiver.enable()
        else:
            cls.elasticrm_receiver.disable()
        
        
        
