"""
`tigres.core.execution.plugin.local`
========================

.. currentmodule:: tigres.core.execution.plugin.local

:platform: Unix, Mac
:synopsis: Execution Local plugin API


.. moduleauthor:: Val Hendrix <vchendrix@lbl.gov>

"""
import os
from copy import copy

from tigres.core.execution import utils as eutils
from tigres.core.execution.plugin import ExecutionPluginBase
from tigres.core.execution.utils import TaskThreadExecution, TaskProcessExecution
from tigres.core.execution.utils import run_command, run_commands
from tigres.core.utils import cpu_count
from tigres.utils import TigresException, State


class ExecutionPluginLocalBase(ExecutionPluginBase):
    """
    Plugin for local (threaded) Tigres Execution

    """

    @classmethod
    def execute_sequence(cls, work, execution_data):
        work_array=work["work"]

        output = None
        copy_os_env = copy(os.environ)
        copy_os_env.update(execution_data['env'])
        if not execution_data or State.DONE not in list(execution_data.keys()):

            commands = []
            for index, w in enumerate(work_array):
                commands.append(eutils.Command(impl=w["inputs"].task.impl_name, args=w["inputs"].values))

            try:
                # run the commands in a pipeline
                output = run_commands(commands, copy_os_env)

            except Exception as err:
                raise TigresException(
                    "Exception caught for execution: Sequence: '{w}'  Error: {e}".format(
                        w=work["name"], e=err))

        if execution_data is not None:
            execution_data[State.DONE] = State.DONE
        return output, State.DONE

    @classmethod
    def execute_task(cls, work, execution_data):
        """ Executes the Task as a function for the given input values on the local machine

        :param name: unique name for this execution
        :param execution_data:
        :param task: the task for execution
        :type task: tigres.type.Task
        :param input_values: the input values for the task execution
        :type input_values: tigres.type.InputValues
        :returns: results and status
        :rtype : tuple


        :Example:

        >>> from tigres.utils import Execution
        >>> from tigres.core.execution import load_plugin
        >>> engine = load_plugin(Execution.LOCAL_THREAD)
        >>> from tigres import InputTypes, InputValues, Task, FUNCTION
        >>> def my_impl(a, b): return "foo {} {}".format(a, b)
        >>> input_type_1 = InputTypes(None, [int, str])
        >>> task1 = Task(None, FUNCTION, my_impl, input_type_1)
        >>> input_values_1 = InputValues(None, [1, 'hello'])
        >>> engine.execute("foo bar", task1, input_values_1,{'env':{}})
        ('foo 1 hello', 'DONE')


        :Example:

        >>> from tigres.utils import Execution
        >>> from tigres.core.execution import load_plugin
        >>> engine = load_plugin(Execution.LOCAL_THREAD)
        >>> from tigres import InputTypes, InputValues, Task, EXECUTABLE
        >>> input_type_1 = InputTypes(None, [str])
        >>> task1 = Task(None, EXECUTABLE, "/bin/echo", input_type_1)
        >>> input_values_1 = InputValues(None, ['world'])
        >>> engine.execute_executable("foo bar", task1, input_values_1,{'env':{}})

        .. note::
            The current implementation will coerce EXECUTABLE  arguments to strings when it builds the
            command for command line execution
        """
        task=work["inputs"].task
        input_values = work["inputs"].values
        output = None
        if not execution_data or State.DONE not in list(execution_data.keys()):
            if task.env:
                execution_data['env'].update(task.env)

            try:
                output = run_command(task.impl_name, input_values, env=execution_data['env'])
                if execution_data is not None:
                    execution_data[State.DONE] = State.DONE
            except Exception as err:
                raise TigresException(
                    "Exception caught for execution '{w}', Task '{t}'. Error: {e}".format(
                        w=work["name"], t=task.name, e=err))

        return output, State.DONE

    @classmethod
    def execute_function(cls, name, task, input_values, execution_data):
        """ Executes the Task as a function for the given input values on the local machine

        :param name: unique name for this execution
        :param execution_data:
        :param task: the task for execution
        :type task: tigres.type.Task
        :param input_values: the input values for the task execution
        :type input_values: tigres.type.InputValues
        :returns: results and status
        :rtype : tuple


        :Example:

        >>> from tigres.utils import Execution
        >>> from tigres.core.execution import load_plugin
        >>> engine = load_plugin(Execution.LOCAL_THREAD)
        >>> from tigres import InputTypes, InputValues, Task, FUNCTION
        >>> def my_impl(a, b): return "foo {} {}".format(a, b)
        >>> input_type_1 = InputTypes(None, [int, str])
        >>> task1 = Task(None, FUNCTION, my_impl, input_type_1)
        >>> input_values_1 = InputValues(None, [1, 'hello'])
        >>> engine.execute_function("foo bar", task1, input_values_1,{'env':{}})
        ('foo 1 hello', 'DONE')

        """
        return ExecutionPluginLocalBase.execute(name, task, input_values, execution_data)

    @classmethod
    def execute_executable(cls, name, task, input_values, execution_data):
        """ Executes the Task as an executable for the given input values on the local machine

        :param task: the task to execute
        :type task: tigres.type.Task
        :param input_values: the input values for the task
        :type input_values: InputValues or list


        :Example:

        >>> from tigres.utils import Execution
        >>> from tigres.core.execution import load_plugin
        >>> engine = load_plugin(Execution.LOCAL_THREAD)
        >>> from tigres import InputTypes, InputValues, Task, EXECUTABLE
        >>> input_type_1 = InputTypes(None, [str])
        >>> task1 = Task(None, EXECUTABLE, "/bin/echo", input_type_1)
        >>> input_values_1 = InputValues(None, ['world'])
        >>> engine.execute_executable("foo bar", task1, input_values_1,{'env':{}})
        ('world', 'DONE')

        .. note::
            The current implementation will coerce all arguments to strings when it builds the
            command for command line execution
        """
        return ExecutionPluginLocalBase.execute(name, task, input_values, execution_data)


class ExecutionPluginLocalProcess(ExecutionPluginLocalBase):
    """
    Plugin for local (threaded) Tigres Execution

    """

    # I hate doing this class variable but it is by far the quickest
    # way for this *prototype* code. It is okay as long as the
    #  methods used in the core code are called
    # correctly. When there is time, this should be refactored.
    task_process = None

    @classmethod
    def startup(cls, run_fn, work_list=None):
        """
        Initializes the parallel engine to accept jobs
        """

        # try:
        #     ExecutionPluginLocalProcess.ps_task_process = TaskProcessExecutionParallelSequential(cls)
        # except Exception as e:
        #     print "Exeception %s" % str(e)            

        ExecutionPluginLocalProcess.task_process = TaskProcessExecution(cls, work=work_list)

    @classmethod
    def shutdown(cls):
        """
        Shutdown the parellel_sequential engine
        """

        # Wait for tasks to finish
        if ExecutionPluginLocalProcess.task_process:
            ExecutionPluginLocalProcess.task_process.join()
            ExecutionPluginLocalProcess.task_process = None

    @classmethod
    def add_work(cls, work):
        """
        add work to be executed
        """

        ExecutionPluginLocalProcess.task_process.submit(work)


class ExecutionPluginLocalThread(ExecutionPluginLocalBase):
    """
    Plugin for local (threaded) Tigres Execution

    """

    # I hate doing this class variable but it is by far the quickest
    # way for this *prototype* code. It is okay as long as the
    # methods used in the core code are called
    # correctly. When there is time, this should be refactored.
    task_threads = None

    @classmethod
    def startup(cls, run_fn):
        """
        Initializes the  engine to accept jobs
        """

        thread_count = cpu_count()

        ExecutionPluginLocalThread.task_threads = TaskThreadExecution(run_fn, thread_count)

    @classmethod
    def shutdown(cls):
        """
        Shutdown the engine
        """

        # Wait for tasks to finish
        if ExecutionPluginLocalThread.task_threads:
            ExecutionPluginLocalThread.task_threads.join()
            ExecutionPluginLocalThread.task_threads = None

    @classmethod
    def add_work(cls, work):
        """
        add work to be executed
        """
        ExecutionPluginLocalThread.task_threads.submit(work)
