"""
.. currentmodule:: tigres.core.execution.utils

:platform: Unix, Mac
:synopsis: Local (Desktop) Tigres Execution with qsub

`tigres.core.execution.utils`
*****************************

Batch queue execution for a Tigres Program


Functions
=========
 * :py:func:`create_executable_command` -  Create the executable command for the given task and input_values
 * :py:func:`run_command` -  Runs the command for the specified task
 * :py:func:`multiprocess_worker` -  Runs multiple worker processes and waits for them to finish.

Classes
=========
 * :py:class:`FOpen` - Executes a function with arguments in a child process.
 * :py:class:`TaskProcessExecution` - Launches multiple processes to execute Tigres tasks in parallel
 * :py:class:`TaskThreadExecution` - Launches multiple threads to execute Tigres tasks in parallel
 * :py:class:`TaskServer` - Task Server for managing Tigres task execution across multiple hosts
 * :py:class:`TaskClient` - Task Client for executing Tigres tasks


.. moduleauthor:: Val Hendrix <vhendrix@lbl.gov>

"""
from __future__ import print_function
import time
import io
import sys
import os
from copy import copy

try:
    import queue
except ImportError:
    import Queue as queue

try:
    xrange
    range = xrange
except NameError:
    pass

import multiprocessing
from multiprocessing.managers import SyncManager
import subprocess
import threading
from tigres.core.utils import get_str, cpu_count
from collections import namedtuple, Callable

try:
    import cPickle as pickle
except ImportError:
    import pickle

try:
    from cloud import cloud
    from cloud.serialization.cloudpickle import dumps
except ImportError:
    from pickle import dumps

from tigres.utils import TigresException, TaskFailure, State

# The named tuple for running commands
Command = namedtuple('Command', ['impl', 'args'])

def create_executable_command(input_values, task):
    """

    :param task: the task for execution
    :type task: tigres.type.Task
    :param input_values: the input values for the task execution
    :type input_values: tigres.type.InputValues
    :return: a command to be run on the command line.
    :rtype: basestring


    """
    if isinstance(task, str):
        cmd = task
    else:
        cmd = task.impl_name

    try:
        # Coerce the input values to strings
        input_values_str = [str(v) for v in input_values]

        if input_values:
            cmd = cmd + ' ' + ' '.join(input_values_str)
    except Exception as e:
        raise TigresException(
            "Exception caught for Task '{}' and command '{}'. Error message: {%}"
            .format(task.name, cmd, e))
    return cmd


def multiprocess_worker(job_queue, result_queue, num_processes, worker):
    """
    Launches the given worker a job and result queue for the specified number
    of processes. Waits until all of the workers are finished.

    :param job_queue: contains the jobs that the worker needs to process
    :param result_queue: empty queue for the worker's results
    :param num_processes: the number of processes to launch
    :param worker: the worker function to pass the the processes
    """
    procs = []
    for _ in range(num_processes):

        p = multiprocessing.Process(
            target=worker,
            args=(job_queue, result_queue))
        procs.append(p)
        p.start()

    # DEBUG:  I think this is not needed?  W/O start run will return
    #         and that worker will end, but so what?
    for p in procs:
        p.join()


def run_commands(commands, env={}):
    """
    Run all the specifed commands as a pipeline. This function accepts
    both function and executable commands

    :param commands: the commands to execute
    :type commands: list of Commands (namedtuple)
    :param env: the environment variables to pass along to the executions
    :type env: dict
    """
    # We have a list of commands, let's chain them with pipes
    num_commands = len(commands)

    # Capture the processes
    processes = []

    # Create all of the processes in the pipeline
    i = 0
    for i in range(num_commands):
        stdin_pipe = None
        command = commands[i]

        # Get the stdin of the previous process
        if ((i - 1) >= 0): stdin_pipe = processes[i - 1].stdout

        run_impl(command, processes, stdin_pipe, env, stdout_pipe=True)

        # Close the stdout of the previous process
        if i > 1:
            prev = processes[i - 1]
            if prev.stdout and not prev.stdout.closed:
                prev.stdout.close()

    # Let's get the output
    out, err = processes[i].communicate()

    if i > 0:
        processes[i - 1].wait() # wait for the previous process to finish

    # close all the stdout
    for p in processes:
        if p.stdout and not p.stdout.closed:
            p.stdout.close()

    # check for errors
    error = None
    # Did the final task in the sequence fail?
    if len(err) > 0:
        error = TigresException("Task '{}' failed with error:{} output:{}"
                              .format(commands[i], err, out))

    # We need to check for other errors and
    # Close the stderr pipes
    errors=[]
    for index, p in enumerate(processes):
        # Look for first error in pipeline
        if not p.stderr.closed:
            perr = p.stderr.read()
            p.stderr.close()  # Gotta close this now so we can check other errors
            if not error and len(perr) > 0:
                errors.append((commands[index],perr))

    if error:
        raise error
    elif len(errors) > 0: # Did something up the chain fail?
        # FIXME - we should figure out how to show the error chain here
        #  for right now just show the first error
        raise TigresException("Command '{} {}' failed with error:{}"
                              .format(errors[0][0].impl, " ".join(errors[0][0].args), errors[0][1]))

    # FIXME what about piping to a file on the final output?
    return get_str(out).rstrip('\n')


def run_impl(command, processes, stdin_pipe, env={}, stdout_pipe=False):
    """
    Execute a command and append it to the process list.  If there,
    is a stdin pipe given pass that along. The command can either be a string
    which will be execute on the command line OR it can be a Callable object.

    :param command: the command to execute
    :type command: Command
    :param processes: a list of processes ( Popen or FOpen )
    :param stdin_pipe: the stdin for the executed command
    :type stdin_pipe: pipe or filelike object
    :param env: the environment variables to pass along to the executions
    :type env: dict


    """
    copy_env = copy(env)
    copy_env.update(os.environ)

    # Are we executing a Python function or an executable on the commandline?
    if isinstance(command.impl, str):
        # Create an executable process
        command_str = create_executable_command(command.args, command.impl)
        processes.append(subprocess.Popen(command_str, stdout=subprocess.PIPE,
                                          stderr=subprocess.PIPE,
                                          stdin=stdin_pipe, env=copy_env, shell=True, close_fds=True))
    elif isinstance(command.impl, Callable) and (stdin_pipe or stdout_pipe):
        # We want to wrap the callable object in a process to accept the standard input
        processes.append(FOpen(command.impl, command.args, stdout=subprocess.PIPE,
                               stderr=subprocess.PIPE,
                               stdin=stdin_pipe, env=copy_env))
    elif isinstance(command.impl, Callable):
        # No standard input for the callable object.  We will just execut the function

        # We copy the python environment and save the old one
        old_env = os.environ

        try:
            # Prepare copied Environment
            os.environ = copy_env

            results = command.impl(*command.args)

            # Return environment back to its original state
            return results
        finally:
            # Return environment back to its original state
            os.environ = old_env
            del copy_env

    else:
        raise TigresException("Invalid Task Implementation '{}': Must be executable or python function".format(
            type(command.impl)))

    # the data was piped
    return None


def run_command(command, args=(), env=None):
    """
    Runs the command for the specified task or list of tasks.
    In the case of a list of commands the output is piped to the next
    command.

    :type env: dict
    :param env:
    :param command: the command line execution
    :type command: list or str
    :return: the command output
    :rtype: str
    """

    command_tuple = Command(command,args)
    processes =[]
    results = run_impl(command_tuple, processes, None, env)
    if not results and isinstance(command, str):
        stdout, stderr = processes[0].communicate()
        if len(stderr) > 0:
            raise TigresException("Command '{}' failed with error:{} output:{}"
                              .format(command, stderr, stdout))

        return get_str(stdout).rstrip('\n')

    return results


def worker(results_queue, input_queue, plugin):
    """
    A worker function
    """

    while 1:
        # get task
        w = input_queue.get()
        if w is None:
            input_queue.task_done()
            results_queue.put(None)
            return  # stop on sentinel value, None

        try:
            try:
                w = pickle.loads(w)
            except Exception as err:
                print(err)
                pass

            results_queue.put(pickle.dumps((w["name"], (None, State.RUN)), protocol=-1))
            results_queue.put(pickle.dumps((w["name"], plugin.execute(w, w["execution_data"])), protocol=-1))

        except Exception as err:
            print(err)
            results_queue.put((w.name, (
                TaskFailure("Exception caught for execution", error=err),
                State.FAIL)))

        # We are done with this task, for now
        input_queue.task_done()


class TaskProcessExecution(object):
    """ Launches multiple processes to execute Tigres tasks in parallel """

    def __init__(self, plugin, work=None):
        """Construct and run.

        :param plugin: Execution plugin
        """

        super(TaskProcessExecution, self).__init__()
        self._queue = multiprocessing.JoinableQueue(0)
        self._num_processes = cpu_count()

        self._num_processes_started  = 0
        self._num_processes_finished = 0
        self._task_processes = []
        self._joined = False
        self._process_manager = multiprocessing.Manager()
        self._results = multiprocessing.JoinableQueue(0)
        self._gather_thread = None
        self._work = {}
        self._work_lock = threading.Lock()

        # ToDo: Debug: This isn't currently being used
        # self._gather_thread_quit = threading.Event()

        # Start the processes
        for _ in range(self._num_processes):
            try:
                t = multiprocessing.Process(target=worker, args=(
                    self._results, self._queue, plugin))
                self._task_processes.append(t)
                t.start()
                self._num_processes_started += 1
            except Exception as err:
                # print(err)
                if t.is_alive():
                    t.terminate()
                while t.is_alive():
                    time.sleep(0.1)

        # if work:
        #     self._submit_work(work)

        # Start the thread to thread to check for results
        self._gather_thread = threading.Thread(target=self._gather_waiting_results)
        self._gather_thread.daemon = True
        self._gather_thread.start()

        if work:
            for w in work:
                self.submit(w)

    def _gather_waiting_results(self):
        """
        Gathers any results that are currently waiting and updates the 
        dictionary of state items. 

        :return: None
        """

        # Gather the results as they come in
        while 1:
            work_results = self._results.get()

            with self._work_lock:
                if work_results is None:
                    self._num_processes_finished += 1
                    self._results.task_done()
                    if self._num_processes_finished == self._num_processes:
                        break  # All of the processes have finished
                else:
                    try:
                        work_results = pickle.loads(work_results)
                    except:
                        pass
                    name = work_results[0]
                    if isinstance(self._work[name], list):
                        self._work[name][-1].results = work_results[1][0]
                        self._work[name][-1].state = work_results[1][1]
                    else:
                        self._work[name].results = work_results[1][0]
                        self._work[name].state = work_results[1][1]
                    self._results.task_done()

    def submit(self, work):
        """
        Submit a unit of work queue
        :param work:
        :type list:
        :return:
        """
        # add the state to the process queue

        with self._work_lock:
            try:
                self._work[work.name]=work

                # can it be pickled?
                self._queue.put(dumps(work.picklable(), protocol=-1))
            except Exception as e:
                # TODO DEBUG FIXME do something better here
                raise TigresException("Could not pickle code in submit. Error %s" % str(e))

    def join(self):
        """Blocks until all items in the internal FIFO queue have been gotten and processed.
        """

        # shut down the gather waiting results thread
        # self._gather_thread_quit.set()
        # self._gather_thread.join()
        # This is not thread safe but that should be OK
        if not self._joined:
            self._joined = True
            # Add sentinel value to stop each thread
            for _ in range(self._num_processes_started):
                self._queue.put(None)
            # Block for threads to finish, storing exceptions
            self._gather_thread.join()

        self._queue.join()


class TaskThreadExecution(object):
    """  Launches multiple threads to execute Tigres tasks in parallel
    """

    def __init__(self, run_fn, num_threads=2):
        """Construct and run.

        :param num_threads: number of threads to run
        :param run_fn: the run function to use
        """
        super(TaskThreadExecution, self).__init__()
        self._queue = queue.Queue(0)
        self._num_threads = num_threads
        self._task_threads = []

        def worker():

            while 1:
                # get task
                work = self._queue.get()
                if work is None:
                    self._queue.task_done()
                    break  # stop on sentinel value, None

                try:
                    run_fn(work)
                except TigresException as err:
                    work.results = TaskFailure(
                        "Task Execution Failure for {}".format(work.name),
                        error=err)
                    work.state = State.FAIL


                # We are done with this task, for now
                self._queue.task_done()

        for _ in range(num_threads):
            t = threading.Thread(target=worker)
            self._task_threads.append(t)
            t.start()

    def submit(self, work):
        """Submits the tasks to the internal FIFO queue for execution.
        """
        self._queue.put(work)

    def join(self):
        """Blocks until all items in the internal FIFO queue have been gotten and processed.
        """
        # Add sentinel value to stop each thread
        for _ in range(self._num_threads):
            self._queue.put(None)
            # Block for threads to finish, storing exceptions
        self._queue.join()


class TaskServerQueueManager(SyncManager):
    pass


class TaskClientQueueManager(SyncManager):
    pass


class TaskServer(object):
    """ Task Server for managing Tigres task execution across multiple hosts """

    def __init__(self, host='localhost', port=0, secret_key=None, work_list=None):
        """

        :param host: task server host name
        :param port: task server port
        :param secret_key: the secret key shared by task and client
        :return:
        """
        job_queue          = multiprocessing.Queue()
        result_queue       = multiprocessing.Queue()
        client_count_queue = multiprocessing.Queue()

        self._address = (host, port)
        self._results = {}
        self._work = {}
        self._work_lock = threading.Lock()
        self._count_work = 0
        self._count_finished_work = 0
        self._gather_thread = None

        TaskServerQueueManager.register('job_queue', callable=lambda: job_queue)
        TaskServerQueueManager.register('result_queue',
                                        callable=lambda: result_queue)
        TaskServerQueueManager.register('client_count_queue',
                                        callable=lambda: client_count_queue)
        self._manager = TaskServerQueueManager(address=self._address,
                                               authkey=secret_key.encode(
                                                   'utf-8'))

        self._manager.start()
        self._job_queue = self._manager.job_queue()

        # Start the thread to check for results
        self._gather_thread = threading.Thread(target=self._gather_results)
        self._gather_thread.daemon = True
        self._gather_thread.start()

        if work_list:
            for w in work_list:
                self.submit(w)

    def submit(self, work):
        # job_queue = self._manager.job_queue()
        job_queue = self._job_queue
        
        with self._work_lock:
            if work is None:
                job_queue.put(None)
                return

            self._count_work += 1
            self._work[work.name] = work

            # can it be pickled?
            job_queue.put(dumps(work.picklable(), protocol=-1))

    def _gather_results(self):
        """
        Gathers the results as they come and updates the dictionary of
        state items
        :return: None
        """

        result_queue = self._manager.result_queue()

        # wait for the job_queue to be empty
        while 1:
            work_results = result_queue.get()
            if work_results is None:
                break

            with self._work_lock:
                self._count_finished_work += 1
                name = work_results[0]
                if name not in self._work:
                    work_results = pickle.loads(work_results)
                    name = work_results[0]
                    if name not in self._work:
                        self.kill()
                        raise Exception("Invalid Work {}: name unknown".format(name))
                if isinstance(self._work[name], list):
                    self._work[name][-1].results = work_results[1][0]
                    self._work[name][-1].state = work_results[1][1]
                else:
                    self._work[name].results = work_results[1][0]
                    self._work[name].state = work_results[1][1]

    def join(self):
        job_queue          = self._manager.job_queue()
        result_queue       = self._manager.result_queue()
        client_count_queue = self._manager.client_count_queue()

        # Gather all the final results
        while 1:
            with self._work_lock:
                if self._count_finished_work >= self._count_work*2:
                    result_queue.put(None)
                    self._gather_thread.join()
                    break
            time.sleep(.2)

        # Close down all the clients that we know about. First gathe all the clients that started.
        client_count = 0
        try:
            while True:
                count = client_count_queue.get(False)
                client_count += count
        except queue.Empty:
            pass

        # Now close down all the clients
        while client_count > 0:
            job_queue.put(None)
            client_count = client_count - 1

        self.kill()

    def kill(self):
        self._manager.shutdown()


class TaskClient(object):
    """
    Task Client for executing Tigres tasks
    """

    # DEBUG: I think there's a problem in that if you do a ps after a workflow ends,
    #        then the worker process stays around for a while. But then they do
    #        do go away nicely after a couple of minutes or so. Possibly this is just
    #        python process/comminucation latency?
    def worker(self, job_queue, result_queue):
        """
        :param job_queue:
        :param result_queue:
        :return:
        """
        work=None
        while True:
            try:
                # Non-blocking get on job_queue
                # May throw empty error
                work = job_queue.get_nowait()
                if work is None:
                    return # Sentinal value
            except queue.Empty:
                try:
                    # This is not reliable
                    # Only if it throws an error
                    # can we be sure that there are
                    # no jobs left in the queue.
                    job_queue.empty()
                    continue
                except:
                    # We are really done
                    return
            except EOFError as eof:
                # This means there are no more jobs
                return
            except IOError as io:
                # This means there are no more jobs
                return
            except ConnectionResetError as cre:
                # This means there are no more jobs
                return

            # Now we have a work object. Let's unpickle it
            # and execute the work
            try:
                try:
                    work = pickle.loads(work)
                    result_queue.put(dumps((work["name"], (None, State.RUN)), protocol=-1))
                    result_queue.put(dumps((work["name"], self._execute_fn(work, work["execution_data"])), protocol=-1))

                except Exception as e:
                    print(e)

            except Exception as err:
                print(err)
                if "name" in work:
                    result_queue.put((work["name"], (
                            TaskFailure("Task Execution Failure for {}".format(work["name"]),
                                        error=err),
                            State.FAIL)))
                else:
                    result_queue.put((None, (
                        TaskFailure("Task Execution Failure for Unknown Task",
                                error=err),
                        State.FAIL)))


    def __init__(self, execute_fn, host='localhost', port=44555,
                 secret_key=None):
        """
        Task Client for executing Tigres tasks

        :param execute_fn:  The execute function
        :param host:
        :param port:
        :param secret_key:
        :return:
        """

        TaskClientQueueManager.register('job_queue')
        TaskClientQueueManager.register('result_queue')
        TaskClientQueueManager.register('client_count_queue')
        self._execute_fn = execute_fn

        self._manager = TaskClientQueueManager(address=(host, port),
                                               authkey=secret_key.encode(
                                                   'utf-8'))
        self._manager.connect()

    def run(self):
        job_queue          = self._manager.job_queue()
        result_queue       = self._manager.result_queue()
        client_count_queue = self._manager.client_count_queue()

        client_count_queue.put(cpu_count())

        multiprocess_worker(job_queue, result_queue,
                            cpu_count(), self.worker, )


class FOpen:
    """
    This class is inspired by subprocess.POpen. The specified function
    with arguments is executed in a child process.
    It is possible to chain the input and output together
    with subprocess.POpen and other FOpen objects.


    :param func: The function to execute
    :type func: function
    :param args: the function arguments
    :type args: list or tuple
    :param stdin: the executed functions standard input file handle. Valid values
                  are subprocess.PIPE, an existing file descriptor, an existing file object and None.
                  subprocess.PIPE tells FOpen to create a new pipe for the child process. None: no
                  redirection will happen
    :type stdin: positive integer
    :param stdout: the executed functions standard output file handle. Valid values
                  are subprocess.PIPE, an existing file descriptor, an existing file object and None.
                  subprocess.PIPE tells FOpen to create a new pipe for the child process. None means no
                  redirection will happen.
    :type stdout: positive integer
    :param stderr: the executed functions standard error file handle. Valid values
                  are subprocess.PIPE, an existing file descriptor, an existing file object and None.
                  subprocess.PIPE tells FOpen to create a new pipe for the child process. None means no
                  redirection will happen.
    :type stderr: positive integer
    """

    def __init__(self, func, args=(), stdin=None, stdout=None, stderr=None, env=None):

        # get the File Handles for stdin, stdout and stderror
        stdin_read, stdin_write = self._get_io(stdin)
        stdout_read, stdout_write = self._get_io(stdout)
        stderr_read, stderr_write = self._get_io(stderr)

        # Prepare the parents end of the pipes
        self.stdin = None
        self.stdout = None
        self.stderr = None

        if stdin_write:
            self.stdin = os.fdopen(stdin_write, 'wb', 0)
        if stdout_read:
            self.stdout = os.fdopen(stdout_read, 'rb', 0)
        if stderr_read:
            self.stderr = os.fdopen(stderr_read, 'rb', 0)

        # The return code
        self.status = None

        # Execute the function in a child processe
        try:
            self._execute_function(func, args, stdin_read, stdin_write,
                                   stdout_read, stdout_write, stderr_read, stderr_write, env=env)
        except Exception as e:
            # Clean up if there is an error
            for f in filter(None, [self.stdin, self.stdout, self.stderr]):
                try:
                    f.close()
                except EnvironmentError: # OS an IO Errors
                    pass  # Ignore errors
            raise e

    def communicate(self):
        """Interact with process: Close stdin.  Read data from
        stdout and stderr, until end-of-file is reached.  Wait for
        process to terminate.

        :return: stdout and stderr.
        :rtype: tuple
        """

        stdout, stderr = "", ""
        if self.stdin and not self.stdin.closed:
            self.stdin.close()
        elif self.stdout and not self.stdout.closed:
            stdout = self.stdout.read()
            self.stdout.close()
        elif self.stderr and not self.stderr.closed:
            stderr = self.stderr.read()
            self.stderr.close()
        self.wait()
        return (stdout, stderr)

    def _execute_function(self, func, args, stdin_read, stdin_write,
                          stdout_read, stdout_write, stderr_read, stderr_write, env=None):
        """
        Execute the function as a child process
        """

        if not args:
            args = []

        self._pid = os.fork()
        if self._pid == 0:
            # We are in the child now
            # We copy the python environment and save the old one
            copy_env = copy(os.environ)
            old_env = os.environ
            try:
                # Prepare copied Environment
                os.environ = copy_env
                if env:
                    os.environ.update(env)

                # Copy the streams to std in,out and err
                if stdin_read:
                    os.dup2(stdin_read, sys.stdin.fileno())
                if stdout_write:
                    os.dup2(stdout_write, sys.stdout.fileno())
                if stderr_write:
                    os.dup2(stderr_write, sys.stderr.fileno())

                # close the parent pipes
                if stdin_write:
                    os.close(stdin_write)
                if stdout_read:
                    os.close(stdout_read)
                if stderr_read:
                    os.close(stderr_read)

                # Close all of the file descriptors
                # that were copied into this process
                self._close_fds()

                # Execute the function with the
                # provided arguments
                sys.argv=[] # we don't want any weird command line args

                output=func(*args)
                if output:
                    print(output,end="")
                sys.stdout.flush()

            except Exception as e:
                # print error to std err:
                print(str(e), file=sys.stderr, end = "")
            finally:
                # Return environment back to its original state
                os.environ = old_env
                del copy_env

            # Exit the process with status n, without calling
            # cleanup handlers, flushing stdio buffers, etc.
            os._exit(0)
        else:
            # We only want to close the pipe ends if they
            # both exist
            if stdin_read and stdin_write:
                os.close(stdin_read)
            if stdout_read and stdout_write:
                os.close(stdout_write)
            if stderr_read and stderr_write:
                os.close(stderr_write)

    def _get_io(self, stream):
        """
        Returns pipe reader and writer.  If subprocess.PIPE is
        encounter the pipes are created
        """
        r, w = None, None
        if not stream:
            pass  # do nothting
        elif stream == subprocess.PIPE:
            try:
                r, w = os.pipe()
            except OSError:
                raise
        elif isinstance(stream, int):
            r = stream
        elif hasattr(stream,"fileno"):
            # Assuming file-like object
            r = stream.fileno()
        else:
            #FIXME what is it?
            pass
        return r, w

    def wait(self):
        """
        Wait for the child process to finish.

        """
        _, self.status = os.waitpid(self._pid, 0)

    def __exit__(self, type, value, traceback):
        if self.stdout and not self.stdout.closed:
            self.stdout.close()
        if self.stderr and not self.stderr.closed:
            self.stderr.close()
        if self.stdin and not self.stdin.closed:
            self.stdin.close()

    def _close_fds(self):
        """
        Close all file descriptions except for stdin stdout and stderr
        """
        try:
            MAXFD = os.sysconf("SC_OPEN_MAX")
        except:
            MAXFD = 256

        try:
            start_fd = 3
            for fd in range(start_fd, MAXFD):
                # Need to check if file descriptor is open
                if os.isatty(fd):
                    os.close(fd)
        except IOError as ioe:
            pass








