"""
`tigres.core.state.program`
================================

.. currentmodule:: tigres.core.state.program

:platform: Unix, Mac
:synopsis: Tigres program state


Classes
-------

 * :class:`Program` - A global Tigres program. Used to maintain program state.
 * :class:`ProgramExecutionError` - Exception class for :class:`Program` errors

.. moduleauthor:: Val Hendrix <vhendrix@lbl.gov>



"""
import time
from uuid import uuid4
import threading
import os
import re

from tigres.core.execution import load_plugin
from tigres.core.execution.elastic import init_elastic
from tigres.core.utils import SingletonMeta
from tigres.core.state.work import WorkUnit, WorkParallel, WorkSequence, \
    Identifier, CURRENT_INDEX, PREVIOUS_INDEX
from tigres.utils import get_new_output_file, State, Execution
from tigres.core.monitoring.log import init, finalize, log_node
from tigres.core.monitoring.common import Keyword, Level, NodeType

from tigres.core.state.recovery import Recovery

class ProgramExecutionError(Exception):
    pass

# Python 2 and 3 metaclass compatibility
_ProgramSingletonMeta = SingletonMeta('ProgramSingletonMeta', (object, ), {})


class Program(_ProgramSingletonMeta):
    """
        Global tigres program. This is a singleton class used
        for managing the state of the tigres program.

        :Example:

        >>> from tigres.core.state.program import Program
        >>> program1 = Program()
        >>> program2 = Program()
        >>> id(program1) == id(program2)
        True

    """

    def __init__(self, name=None, log_dest=None,
                 execution=Execution.LOCAL_THREAD, recover=False, recovery_log=False, env=None, log_meta=None,**kwargs):
        """
        Initializes a new Tigres Program

        :param self:
        :param execution: The execution plugin to use. Default: :class:`Execution.LOCAL_THREAD`
        :param env: environment variable for task execution
        :type env: dict
        :param kwargs:
        :param name: Workflow name. If not given, a random one will be chosen.
        :type name: str
        :param log_dest: Log destination, passed as `dest` to `monitoring.api.init()`
        :param log_meta: dictionary containing extra metatdata to log for this workflow
        :type log_meta: dict
        :param log_*: Additional keywords to pass to :func:`monitoring.log.init()`, with the `log_` prefix stripped
        :return: program
        :rtype: Program

        """

        self._lock = threading.Lock()
        if name is None:
            # Autogenerate a name
            name = "Tigres{:x}".format(int(time.time()))

        # All Tasks are registered here
        self._tigres_objects = {}
        self._env = env

        # All state is registered here
        self._work = {}

        self._name = name
        self._identifier = str(uuid4())

        # init monitoring
        monitoring_kw = {k[4:]: v for k, v in kwargs.items() if
                         k.startswith("log_") and k != "log_dest"}

        monitoring_kw['recover'] = True

        load_plugin(execution)

        self._log_dest = log_dest
        if log_dest is None:

            #!
            """
            recover = False
            # Log message: "Recovery failed, because no log_dest provided"
            """
            #!

            self._log_dest = get_new_output_file(
                basename='tigres-{}'.format(name))

        init(self._log_dest, self.name, self._identifier, metadata=log_meta,**monitoring_kw)
        self._log(Level.INFO, "start", state=State.RUN)
        self._log(Level.INFO, Keyword.pfx + "init_program",
                  message="initializing program")

        # This must happen after the logging is initialized
        # We might want to revisit whether we need to log
        # every time a name is registered with the program
        self._root_sequence_work = self.register_sequence_work(name)



        """ Logic for initializing Recovery object """
        log_dir = os.path.dirname(self._log_dest)
        if recovery_log is not False and recovery_log is not True: # recovery_log is a program id
            self._recovery = Recovery(log_dir, recover_prgm_id=recovery_log, recover=recover, recovery_log=recovery_log)
        else:
            self._recovery = Recovery(log_dir, recover=recover, recovery_log=recovery_log)
        
        init_elastic()

    @staticmethod
    def _get_keys_by_name(name, dictionary):
        like_names = [key for key in list(dictionary.keys()) if
                      key.name == name]
        return like_names

    ### Not Currently Used !!!
    @staticmethod
    def isparallel_sequential(work):
        """determine if the work looks like a parallel_sequential"""

        if len(work) == 0:
            return False
        else:
            for w in work:
                if not isinstance(current_work, WorkSequence):
                    return False

        return True


    def _get_unique_identifier(self, name, dictionary):
        # Are there any names already registered?
        like_names = self._get_keys_by_name(name, dictionary)
        len_names = len(like_names)
        # This name exists, create a new WorkId
        identifier = Identifier(name, len_names)
        return identifier

    def _log(self, level, event, nodetype=NodeType.PROGRAM, **kwargs):
        """
        Log messages for the program
        :param level:
        :param status:
        :param kwargs:
        :return:
        """
        log_node(level, self._name, event=event, nodetype=nodetype,
                 **kwargs)

    def _register_work(self, name, work_class):
        """
        base register state method

        :param name: name of state
        :param work_class: the state class to create
        :type work_class: Work, ParallelWork, SequenceWork
        :return: new state object
        """
        if not name and work_class:
            raise ProgramExecutionError(
                "Work Registration error: A name must be specified for class %s" % work_class)
        elif not name:
            raise ProgramExecutionError(
                "Work Registration error: A name and state class must be specified to register state")
        if not work_class:
            raise ProgramExecutionError(
                "Work Registration error: A Work class must be specified for %s" % name)
        parent = None

        with self._lock:
            if hasattr(self, '_root_sequence_work') and len(self._root_sequence_work) > 0:
                if work_class is WorkUnit:
                    parent = self._root_sequence_work[CURRENT_INDEX]
            elif hasattr(self, '_root_sequence_work'):
                parent = self._root_sequence_work

            work_id = self._get_unique_identifier(name, self._work)
            work = work_class(parent, work_id)
            self._work[work_id] = work
            self._log(Level.INFO, "register",
                      message="registering {} '{}' ".format(type(work).__name__, work.name))
            return work

    @staticmethod
    def clear():
        """
        Clear the program
        :return: None
        """
        try:
            if hasattr(Program, '_instance') and Program._instance:
                Program()._log(Level.INFO, "end", state=State.DONE)
                finalize()
        finally:
            Program._instance=None

    @property
    def env(self):
        return self._env

    @property
    def identifier(self):
        return self._identifier

    @property
    def name(self):
        with self._lock:
            if hasattr(self, '_root_sequence_work'):
                return self._root_sequence_work.name
            else:
                return self._name
    @property
    def recovery(self):
        return self._recovery


    def previous_work_params(self, work_index = None):
        """
        This method is like the property previous_work except it can accept
        parameters. Up until parallel_sequential was added to Tigres, PREVIOUS
        could be evaluated without passing parameters to preivous_work. This
        was because at any given time there was only one work being executed
        at any one time. parallel_sequential has several sequences being
        executed at the same time and so parameters need to be passed in
        to specify the particular work unit.

        At some point, this code should be more integrated into what is
        being done by the property previous_work.
        """

        with self._lock:
            current_work = self._root_sequence_work[CURRENT_INDEX]

            # DEBUG
            # current_work = current_work[1] # or 0
            # print "    prev_work: is sequence %s, %s, %s" % (is_sequence, current_work.name, type(current_work))
            # print "    work len %d" % len(current_work)
            # return current_work[0]
            # END DEBUG

            if work_index is not None:
                return current_work[work_index][PREVIOUS_INDEX]
            else:
                return self.previous_work




    @property
    def previous_work(self):
        """
        :param work_index: work_index is an index into the current work. It
        is used when the current work contains multiple active embedded 
        instances of work. If work_index is specified, the sequence algorithm
        is used to find the correct frame regardless of the type of current work.

        Return the :class:`WorkBase` from the previous execution.
        """

        # THIS IS WHAT NEEDS TO CHANGE! (I THINK);  HERE WE NEED TO MIX THE 
        # SYMANTICS OF SEQUENCE AND PARALLEL. NEED TO MAKE SURE TO HAND
        # BACK THE RIGHT SEQUENCE. I THINK SPLIT AND MERGE ARE SIMILIAR?

        # I think the problem might be that this is the first time more than
        # one Work is being updated at the same time, I need to return
        # something in it? MAYBE IT IS A PROBLEM THAT I NEED TO KNOW THE
        # LANE TO RETURN?

        # Might need a new type of work ParallelSequential just to key
        # off of.

        # Somewhere, I'm going to need to store the lane where I can get at it.

        # Key: Up until now, PREVIOUS always meant the same thing. Now, the 
        #      meaning of PREVIOUS is dependent on the task it is called from!
        #      There are many "PREVIOUS" tasks at any one time.
        #  --> Could hack it by sending in a special previous that is used
        #      just internally. It would be entered by _WorkPSGenerator. It would
        #      contain the lane number. This could be accomplished several ways.
        #      a special syntax, another name PS_PREVIOUS, a hidden variable to
        #      PREVIOUS.  Right now, this seems like the best choice. We only 
        #      know which lane each task belongs to at the top layer, so it needs
        #      to be saved there.
        #      PREVIOUS.lane[0]

        with self._lock:
            current_work = self._root_sequence_work[CURRENT_INDEX]
            is_sequence = isinstance(current_work, WorkSequence)

            # DEBUG
            # current_work = current_work[1] # or 0
            # print "    prev_work: is sequence %s, %s, %s" % (is_sequence, current_work.name, type(current_work))
            # print "    work len %d" % len(current_work)

            # return current_work[0]
            # END DEBUG

            if len(current_work) > 1 and is_sequence:
                # The current state is a sequence, and there is previous task
                # available
                return self._root_sequence_work[CURRENT_INDEX][PREVIOUS_INDEX]
            elif len(self._root_sequence_work) > 1:
                # The current state is either parallel or it is the first
                # task in a sequence.  Therefore, the results of the
                # previous template is returned
                return self._root_sequence_work[PREVIOUS_INDEX]
            else:
                # There are not previous tasks to retrieve.
                raise ValueError("There is no previous state available.")

    def get_work_by_name(self, name):
        """
        Get all the state objects that have the specified name in their
        WorkId.

        :param name:
        :return: Work objects with the name
        :rtype: list
        """
        with self._lock:
            return self._get_keys_by_name(name, self._work)

    @property
    def root_work(self):
        """
        Root state object
        :return: the root state object
        :rtype: tigres.core.state.WorkSequence
        """
        with self._lock:
            return self._root_sequence_work

    def register(self, item):
        """

        :param item: The item to register
        :return: the new identifier
        """
        with self._lock:
            identifier = None
            if hasattr(item, "_identifier"):
                identifier = self._get_unique_identifier(item.name,
                                                         self._tigres_objects)
                self._tigres_objects[identifier] = item
                name = identifier.name
                if identifier.index > 0:
                    name += "-{}".format(identifier.index)
                self._log(Level.INFO, "register",
                          message="registering identifier '{}' for {} with name '{}' ".format(
                              name, type(item).__name__,
                              item.name))
            return identifier

    def register_work(self, work_inputs): #! Consider making default State.NEW
        """
        Register the given WorkInputs as new state.
        :param work_inputs: a new Work object with a unique WorkId
        :return: a new state object
        :rtype: tigres.core.state.WorkUnit

        """
        if not work_inputs:
            raise ProgramExecutionError(
                "Work inputs for %s are missing" % work_inputs.task.name)
        # print "---> %s" % (str(work_inputs.task))
        work = self._register_work(work_inputs.task.unique_name, WorkUnit)
        work.inputs = work_inputs
        work.execution_data.setdefault('env', {})
        if self._env:
            work.execution_data['env'].update(self._env)
        return work

    def register_parallel_work(self, name):
        """
        Register the given name as new ParallelWork.
        :param name: user specified name
        :return: new state object
        :rtype: tigres.core.state.WorkParallel
        """
        return self._register_work(name, WorkParallel)

    def register_parallel_sequential_work(self, name):
        """
        Register the given name as new ParallelSequentialWork.
        :param name: user specified name
        :return: new state object
        :rtype: tigres.core.state.WorkParallel
        """
        return self._register_work(name, WorkParallelSequential)

    def register_sequence_work(self, name):
        """
        Register the given name as new SequenceWork
        :param name: user specified name
        :return: new state object
        :rtype: tigres.core.state.WorkSequence
        """
        return self._register_work(name, WorkSequence)
