
"""
`tigres.state.recovery`
**********************

.. currentmodule:: tigres.state.recovery

:platform: Unix, Mac
:synopsis: resumes a workflow where it last left off

Users can turn on this module's functionality by setting parameter "recover" when calling the user-level API method start()

Classes
=======
 * :class:`Recovery` - Provides functions for recovering task statuses, inputs, and outputs. Enables internal filtering of
                        previously completed tasks on workflow re-runs, following task failures.

Functions
=========
 
 * :func:`resolve_status` - Returns recovery status of WorkUnit from records
 * :func:`resolve_work_outputs` - Look up given work's logged outputs, and return deserialized results



.. moduleauthor:: James Fox <jsfox@lbl.gov>
"""

import json
import os
from tigres.core.monitoring.common import Keyword
from tigres.core.monitoring import log
from tigres.core.monitoring.log import log_node, _init_logger
from tigres.utils import State
from tigres.core.utils import JSONObjectsUtil
import threading

DATA_RECOVERY_DIR = "{}-data-logs"


class Recovery(object):

    __DATA_DIR_LOCK = threading.RLock()

    """
    Recovery class handles data (input) and task recovery, from log(s).
    Filters out work that don't need to be rerun, when recovering from a failure.
    
    """

    def __init__(self, log_dir, recover_prgm_id=None, recover=False, recovery_log=True):
        self.recover = recover
        self.recovery_log = recovery_log
        self.curr_prgm_id = log._program_uuid
        self.recover_prgm_id = recover_prgm_id
        if recover and not recover_prgm_id:
            self.recover_prgm_id = _most_recent_program_id()
        self.log_dir = log_dir

    def resolve_status(self, work_unit):
        """
        Returns recovery status of WorkUnit from records

        :param rec: Record
        :return: State.RECOVERED, RERUN, or NEW
        """
        log_status = self._get_log_status(work_unit)

        if not log_status:
            return State.NEW

        last_status = log_status.state
        if last_status == State.DONE or last_status == State.RECOVERED:
            return State.RECOVERED
        elif last_status == State.FAIL:
            return State.RERUN
        else:
            return State.NEW


    def resolve_work_outputs(self, work_unit):
        """
        Look up recorded outputs for given task, and return deserialized results

        :param WorkUnit:
        :return: Object or None

        """
        task_log = self.getTaskLog(work_unit.name, self.recover_prgm_id)
        log_status = self._get_log_status(work_unit, task_log=task_log)
        task_output = log_status.output if log_status != None else None
        return JSONObjectsUtil.deserialize(task_output) if task_output != None else None


    def _get_log_status(self, work_unit, task_log=None):
        """
        Checks log and returns corresponding LogStatus of the (first) task with matching name

        """
        from tigres.core.task import Task
        assert type(work_unit.inputs[0]) == Task

        if not self.recover_prgm_id:
            return None

        if task_log:
            log_status_list = log.check('task', program_id=self.recover_prgm_id, names=["^" + work_unit.name], task_log=task_log)
        else:
            log_status_list = log.check('task', program_id=self.recover_prgm_id, names=["^" + work_unit.name])

        if not log_status_list:
            return None
        else:
            return log_status_list[0]

    def _log_task_output(self, log_level, workInst, log_args):
        """
        Records task output to its log
        """
        task_log = self.getTaskLog(workInst.name, self.curr_prgm_id)
        log_args['task_log'] = task_log
        try:
            log_args[Keyword.OUTPUT] = json.dumps(workInst.results)
        except TypeError:  # non-builtin python obj
            log_args[Keyword.OUTPUT] = JSONObjectsUtil.serialize(workInst.results)
        finally:
            log_node(log_level, workInst.name, nodetype=workInst.get_type_name(), **log_args)

    def getTaskLog(self, name, program_id):
        """
        Fetches or creates the task log corresponding to specified task name and program id.
        """
        dirPath = os.path.join(self.log_dir, DATA_RECOVERY_DIR.format(program_id))

        # Only one should create the log directory
        with Recovery.__DATA_DIR_LOCK:
            if not os.path.isdir(dirPath):
                os.mkdir(dirPath)

        name = cleanString(name) # necessary to avoid truncation of task names that contain certain special characters.
        logName = program_id + "-{}.log".format(name)
        logPath = os.path.join(dirPath, logName)
        task_log = _init_logger(logPath, logName)
        return task_log

CHAR_BLACKLIST = set(['?', ';', '#', '/'])

def cleanString(string, black_list=CHAR_BLACKLIST, replacement='+'):
    """
    Replaces any blacklisted characters in string by the replacement character.

    """
    stringChars = list(string)
    filteredChars = [c if c not in black_list else replacement for c in stringChars]
    return ''.join(filteredChars)

def _most_recent_program_id():
    """
    fetches most recent program id from log
    :return: prgm_id

    """
    prgm_id = None

    for rec in log.query(['{} != user'.format(Keyword.EVENT), '{} = {}'.format(Keyword.EVENT, "RUN"), \
                          '{} ~ {}'.format(Keyword.NODETYPE, "template.*"), '{} ~ {}'.format(Keyword.PROGRAM_UID, '.*')]):
        prgm_id = rec[Keyword.PROGRAM_UID]

    return prgm_id



