"""
.. currentmodule:: tigres.core.utils

:platform: Unix, Mac
:synopsis: The internal task api

`tigres.core.utils`
*******************

Core utilities for Tigres

Functions
=========
 * :py:func:`cpu_count` - tries to determine the physical cpu count then returns logical of those methods fail
 * :py:func:`get_free_port` - returns a port free on the current host
 * :py:func:`get_metaclass` - returns the meta class for the given class
 * :py:func:`get_str` - Python version 2/3 compatible string conversion.


Classes
=========
 * :py:class:`SingletonMeta` -  Singleton Metaclass for making a singleton class
 * :py:class:`TigresInternalException` - Is this used?
 * :py:class:`JSONObjectsUtil` - Provides methods for JSON serialization and deserialization (into original objects)

.. moduleauthor:: Val Hendrix <vhendrix@lbl.gov>

"""
import socket
import sys
import os
import re
import subprocess
import json
import inspect

def get_metaclass(cls, name):
    return cls(name, (object, ), {})


def get_str(s):
    if sys.version_info >= (3, 0, 0):
        # for Python 3
        if isinstance(s, bytes):
            s = s.decode('ascii')  # or  s = str(s)[2:-1]
    else:
        # for Python 2
        if isinstance(s, unicode):
            s = str(s)
    return s


def get_free_port():
    """
    Get a free port
    :return:
    """
    s = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
    s.bind(('localhost', 0))
    _, port = s.getsockname()
    s.close()
    return port

def cpu_count():
    """
    Tries to determine the physical cpu count then returns logical/virtual if those methods fail.
    :return: the number of cpus
    """

    # lscpu
    try:
        lscpu = subprocess.Popen(['lscpu'],
                                 stdout=subprocess.PIPE,stderr=subprocess.PIPE)
        out = lscpu.communicate()[0]
        out = str(out).split("\n")
        res = 0
        for line in out:
            line = line.rstrip()
            if re.search(".*ocket", line):
                if res == 0:
                    res = 1
                res *= int(line.split(":")[1])

        if res > 0:
            return res
    except (OSError, ValueError):
        pass


    # BSD
    try:
        sysctl = subprocess.Popen(['sysctl', '-n', 'hw.physicalcpu'],
                                  stdout=subprocess.PIPE,stderr=subprocess.PIPE)
        out = sysctl.communicate()[0]
        res = int(out)

        if res > 0:
            return res
    except (OSError, ValueError):
        pass

    # Solaris
    try:
        pseudoDevices = os.listdir('/devices/pseudo/')
        res = 0
        for pd in pseudoDevices:
            if re.match(r'^cpuid@[0-9]+$', pd):
                res += 1

        if res > 0:
            return res
    except OSError:
        pass

    # Other UNIXes (heuristic)

    try:
        try:
            dmesg = open('/var/run/dmesg.boot').read()
            dmesg = str(dmesg)
        except IOError:
            dmesgProcess = subprocess.Popen(['dmesg'], stdout=subprocess.PIPE,
                    stderr=subprocess.PIPE)
            dmesg = dmesgProcess.communicate()[0]
            dmesg = str(dmesg)

        res = 0
        while '\ncpu' + str(res) + ':' in dmesg:
            res += 1

        if res > 0:
            return res

    except OSError:
        pass

    # Python 2.6+
    try:
        import multiprocessing

        return multiprocessing.cpu_count()
    except (ImportError, NotImplementedError):
        pass

    raise Exception('Can not determine number of CPUs on this system')


# noinspection PyAttributeOutsideInit,PyArgumentList
class SingletonMeta(type):
    """ Singleton Metaclass for making a singleton class """

    def __call__(cls, *args, **kwargs):
        """

        :param cls: Class object that is parent of this meta class
        :param args: positional arguments
        :param kwargs: keyword arguments
        :return: The singular instance
        """

        if '_instance' not in cls.__dict__ or cls._instance is None:
            cls._instance = super(SingletonMeta, cls).__call__(*args, **kwargs)
        return cls._instance

class JSONObjectsUtil(object):
    """
    Methods dealing with JSON serialization and deserialization of objects.
    """

    METADATA = 'tg_obj_meta'
    DATA = 'tg_obj_data'
    MODULE_KEY = 'module_name'
    CLASS_KEY = 'class_name'

    JSON_SERIALIZABLE_TYPES = [str, int, float, bool, type(None)] # according to 'Dive Into Python 3.
    # Note: list & dict are serializable unless they contain non-serializable elements
    try:
        JSON_SERIALIZABLE_TYPES.append(unicode) # if python 2
    except NameError:
        pass

    @staticmethod
    def serialize(item):
        """
        Serializes item to JSON, after first making the item serializable using custom logic
        :return: JSON string

        """
        return json.dumps(JSONObjectsUtil.get_json_serializable(item))

    @staticmethod
    def get_json_serializable(item):
        """
        Converts item to a JSON-serializable

        """
        if type(item) in JSONObjectsUtil.JSON_SERIALIZABLE_TYPES:  # base case
            return item
        elif type(item) is list: # list case
            serializable_list = []
            for elt in item:
                serializable_list.append(JSONObjectsUtil.get_json_serializable(elt))
            return serializable_list
        elif type(item) is dict: # dictionary case
            serializable_dict = dict()
            for k, v in item.items():
                serializable_dict[k] = JSONObjectsUtil.get_json_serializable(v)
            return serializable_dict
        else:  # non-builtin python object case
            return JSONObjectsUtil.handle_nonserializable_json(item)  # mutual recursion
                                                                      # with handle_nonserializable_json

    @staticmethod
    def handle_nonserializable_json(item):
        """
        Try to 'massage' JSON-unserializable item into serializable form
        """

        output = None
        if not hasattr(item, '__dict__'):
            raise TigresInternalException("Deserialization of {} not supported".format(type(item)))
        else:
            obj_data = {}
            for key, val in item.__dict__.items():
                obj_data[key] = JSONObjectsUtil.get_json_serializable(val)  # mutual recursion
            output = {
                JSONObjectsUtil.DATA: obj_data,
                JSONObjectsUtil.METADATA: {
                    JSONObjectsUtil.MODULE_KEY: inspect.getmodule(item).__name__,
                    JSONObjectsUtil.CLASS_KEY: item.__class__.__name__
                }
            }
        return output

    @staticmethod
    def deserialize(item):
        """
        Reconstructs an object from its deserialized JSON. Designed to take results of JSONObjectsUtil.picklable() as input
        NOTE: json library treats keys as strings during serialization, and remain so through deserialization

        :param item: JSON string
        :return: obj
        """
        return JSONObjectsUtil.unpack(json.loads(item))

    @staticmethod
    def unpack(loaded_item):
        """
        Restores object description to its proper class instance.

        :param loaded_item: result of json.loads()
        :return: obj

        """
        item_type = type(loaded_item)
        if item_type in JSONObjectsUtil.JSON_SERIALIZABLE_TYPES:
            return loaded_item
        elif item_type is list:
            for i, elt in enumerate(loaded_item):
                loaded_item[i] = JSONObjectsUtil.unpack(elt)
            return loaded_item
        elif item_type is dict and not JSONObjectsUtil.METADATA in loaded_item.keys() or not JSONObjectsUtil.DATA in loaded_item.keys():
            for k, v in loaded_item.items():
                loaded_item[k] = JSONObjectsUtil.unpack(v)
            return loaded_item
        elif item_type is dict and JSONObjectsUtil.METADATA in loaded_item.keys() and JSONObjectsUtil.DATA in loaded_item.keys(): # there is info for re-creating this object
            obj_dict = loaded_item
            obj_metadata, obj_data = obj_dict[JSONObjectsUtil.METADATA], obj_dict[JSONObjectsUtil.DATA]
            obj_module = sys.modules[obj_metadata[JSONObjectsUtil.MODULE_KEY]]
            cls_name = obj_metadata[JSONObjectsUtil.CLASS_KEY]
            obj_cls_info = list(filter(lambda pair: pair[0] == cls_name, inspect.getmembers(obj_module, inspect.isclass)))
            if obj_cls_info:
                obj_cls = obj_cls_info[0][1] # get 1st (and hopefully only item) of list, which is a tuple, and fetch its 2nd entry
                restored = JSONObjectsUtil.create_obj(obj_cls, obj_data)
                return restored
            else:
                raise TigresInternalException("Class '{}' could not be found; deserialization failed".format(cls_name))
        else:
            raise TigresInternalException("Unrecognized type '{}'".format(item_type))

    #NOTE: __new__ won't work for 'old style' classes (< Python 2.1)
    @staticmethod
    def create_obj(obj_class, inst_info):
        """
        Create and initialize object according to provided info

        :param obj_class: object's class
        :param inst_info: instance variables info
        :type dict
        :return: obj

        """
        new_obj = obj_class.__new__(obj_class)
        for key, val in inst_info.items():
            deserialized_val = JSONObjectsUtil.unpack(val) # mutual recursion with unpack()
            setattr(new_obj, key, deserialized_val)
        return new_obj


class TigresInternalException(Exception):
    pass
